from functools import partial
from random import choice, randint, random, sample
from typing import TYPE_CHECKING, Optional

from golem.core.adapter import register_native
from golem.core.dag.graph import ReconnectType
from golem.core.dag.graph_node import GraphNode
from golem.core.dag.graph_utils import distance_to_root_level, distance_to_primary_level, graph_has_cycle
from golem.core.optimisers.advisor import RemoveType
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.optimisers.opt_node_factory import OptNodeFactory
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, AlgorithmParameters
from golem.utilities.data_structures import ComparableEnum as Enum

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters


class MutationStrengthEnum(Enum):
    weak = 0.2
    mean = 1.0
    strong = 5.0


class MutationTypesEnum(Enum):
    simple = 'simple'
    growth = 'growth'
    local_growth = 'local_growth'
    tree_growth = 'tree_growth'
    reduce = 'reduce'
    single_add = 'single_add'
    single_change = 'single_change'
    single_drop = 'single_drop'
    single_edge = 'single_edge'

    none = 'none'


def get_mutation_prob(mut_id: MutationStrengthEnum, node: Optional[GraphNode],
                      default_mutation_prob: float = 0.7) -> float:
    """ Function returns mutation probability for certain node in the graph

    :param mut_id: MutationStrengthEnum mean weak or strong mutation
    :param node: root node of the graph
    :param default_mutation_prob: mutation probability used when mutation_id is invalid or graph has cycles
    :return mutation_prob: mutation probability
    """
    mutation_prob = default_mutation_prob
    graph_cycled = node is None
    if node:
        graph_cycled = distance_to_primary_level(node) < 0
    if mut_id in list(MutationStrengthEnum) and not graph_cycled:
        mutation_strength = mut_id.value
        mutation_prob = mutation_strength / (distance_to_primary_level(node) + 1)
    return mutation_prob


@register_native
def simple_mutation(graph: OptGraph,
                    requirements: GraphRequirements,
                    graph_gen_params: GraphGenerationParams,
                    parameters: 'GPAlgorithmParameters'
                    ) -> OptGraph:
    """
    This type of mutation is passed over all nodes of the tree started from the root node and changes
    nodes’ operations with probability - 'node mutation probability'
    which is initialised inside the function

    :param graph: graph to mutate
    """

    exchange_node = graph_gen_params.node_factory.exchange_node
    visited_nodes = set()

    def replace_node_to_random_recursive(node: OptNode) -> OptGraph:
        if node not in visited_nodes and random() < node_mutation_probability:
            new_node = exchange_node(node)
            if new_node:
                graph.update_node(node, new_node)
            # removed node must not be visited because it's outdated
            visited_nodes.add(node)
            # new node must not mutated if encountered further during traverse
            visited_nodes.add(new_node)
            for parent in node.nodes_from:
                replace_node_to_random_recursive(parent)

    root_nodes = graph.root_nodes()
    root_node = choice(root_nodes) if root_nodes else None
    node_mutation_probability = get_mutation_prob(mut_id=parameters.mutation_strength,
                                                  node=root_node)

    root_node = root_node or choice(graph.nodes)
    replace_node_to_random_recursive(root_node)

    return graph


@register_native
def single_edge_mutation(graph: OptGraph,
                         requirements: GraphRequirements,
                         graph_gen_params: GraphGenerationParams,
                         parameters: 'GPAlgorithmParameters'
                         ) -> OptGraph:
    """
    This mutation adds new edge between two random nodes in graph.

    :param graph: graph to mutate
    """

    def nodes_not_cycling(source_node: OptNode, target_node: OptNode):
        parents = source_node.nodes_from
        while parents:
            if target_node not in parents:
                grandparents = []
                for parent in parents:
                    grandparents.extend(parent.nodes_from)
                parents = grandparents
            else:
                return False
        return True

    for _ in range(parameters.max_num_of_operator_attempts):
        if len(graph.nodes) < 2 or graph.depth > requirements.max_depth:
            return graph

        source_node, target_node = sample(graph.nodes, 2)
        if source_node not in target_node.nodes_from:
            if graph_has_cycle(graph):
                graph.connect_nodes(source_node, target_node)
                break
            else:
                if nodes_not_cycling(source_node, target_node):
                    graph.connect_nodes(source_node, target_node)
                    break
    return graph


@register_native
def add_intermediate_node(graph: OptGraph,
                          node_to_mutate: OptNode,
                          node_factory: OptNodeFactory) -> OptGraph:
    # add between node and parent
    new_node = node_factory.get_parent_node(node_to_mutate, is_primary=False)
    if not new_node:
        return graph

    # rewire old children to new parent
    new_node.nodes_from = node_to_mutate.nodes_from
    node_to_mutate.nodes_from = [new_node]

    # add new node to graph
    graph.add_node(new_node)
    return graph


@register_native
def add_separate_parent_node(graph: OptGraph,
                             node_to_mutate: OptNode,
                             node_factory: OptNodeFactory) -> OptGraph:
    # add as separate parent
    new_node = node_factory.get_parent_node(node_to_mutate, is_primary=True)
    if not new_node:
        # there is no possible operators
        return graph
    if node_to_mutate.nodes_from:
        node_to_mutate.nodes_from.append(new_node)
    else:
        node_to_mutate.nodes_from = [new_node]
    graph.nodes.append(new_node)
    return graph


@register_native
def add_as_child(graph: OptGraph,
                 node_to_mutate: OptNode,
                 node_factory: OptNodeFactory) -> OptGraph:
    # add as child
    old_node_children = graph.node_children(node_to_mutate)
    new_node_child = choice(old_node_children) if old_node_children else None
    new_node = node_factory.get_node(is_primary=False)
    if not new_node:
        return graph
    graph.add_node(new_node)
    graph.connect_nodes(node_parent=node_to_mutate, node_child=new_node)
    if new_node_child:
        graph.connect_nodes(node_parent=new_node, node_child=new_node_child)
        graph.disconnect_nodes(node_parent=node_to_mutate, node_child=new_node_child,
                               clean_up_leftovers=True)

    return graph


@register_native
def single_add_mutation(graph: OptGraph,
                        requirements: GraphRequirements,
                        graph_gen_params: GraphGenerationParams,
                        parameters: AlgorithmParameters
                        ) -> OptGraph:
    """
    Add new node between two sequential existing modes

    :param graph: graph to mutate
    """

    if graph.depth >= requirements.max_depth:
        # add mutation is not possible
        return graph

    node_to_mutate = choice(graph.nodes)

    single_add_strategies = [add_as_child, add_separate_parent_node]
    if node_to_mutate.nodes_from:
        single_add_strategies.append(add_intermediate_node)
    strategy = choice(single_add_strategies)

    result = strategy(graph, node_to_mutate, graph_gen_params.node_factory)
    return result


@register_native
def single_change_mutation(graph: OptGraph,
                           requirements: GraphRequirements,
                           graph_gen_params: GraphGenerationParams,
                           parameters: AlgorithmParameters
                           ) -> OptGraph:
    """
    Change node between two sequential existing modes.

    :param graph: graph to mutate
    """
    node = choice(graph.nodes)
    new_node = graph_gen_params.node_factory.exchange_node(node)
    if not new_node:
        return graph
    graph.update_node(node, new_node)
    return graph


@register_native
def single_drop_mutation(graph: OptGraph,
                         requirements: GraphRequirements,
                         graph_gen_params: GraphGenerationParams,
                         parameters: AlgorithmParameters
                         ) -> OptGraph:
    """
    Drop single node from graph.

    :param graph: graph to mutate
    """
    if len(graph.nodes) < 2:
        return graph
    node_to_del = choice(graph.nodes)
    node_name = node_to_del.name
    removal_type = graph_gen_params.advisor.can_be_removed(node_to_del)
    if removal_type == RemoveType.with_direct_children:
        # TODO refactor workaround with data_source
        graph.delete_node(node_to_del)
        nodes_to_delete = \
            [n for n in graph.nodes
             if n.descriptive_id.count('data_source') == 1 and node_name in n.descriptive_id]
        for child_node in nodes_to_delete:
            graph.delete_node(child_node, reconnect=ReconnectType.all)
    elif removal_type == RemoveType.with_parents:
        graph.delete_subtree(node_to_del)
    elif removal_type == RemoveType.node_rewire:
        graph.delete_node(node_to_del, reconnect=ReconnectType.all)
    elif removal_type == RemoveType.node_only:
        graph.delete_node(node_to_del, reconnect=ReconnectType.none)
    elif removal_type == RemoveType.forbidden:
        pass
    else:
        raise ValueError("Unknown advice (RemoveType) returned by Advisor ")
    return graph


@register_native
def tree_growth(graph: OptGraph,
                requirements: GraphRequirements,
                graph_gen_params: GraphGenerationParams,
                parameters: AlgorithmParameters,
                local_growth: bool = True) -> OptGraph:
    """
    This mutation selects a random node in a tree, generates new subtree,
    and replaces the selected node's subtree.

    :param graph: graph to mutate
    :param local_growth: if true then maximal depth of new subtree equals depth of tree located in
    selected random node, if false then previous depth of selected node doesn't affect to
    new subtree depth, maximal depth of new subtree just should satisfy depth constraint in parent tree
    """
    node_from_graph = choice(graph.nodes)
    if local_growth:
        max_depth = distance_to_primary_level(node_from_graph)
        is_primary_node_selected = (not node_from_graph.nodes_from) or (node_from_graph != graph.root_node and
                                                                        randint(0, 1))
    else:
        max_depth = requirements.max_depth - distance_to_root_level(graph, node_from_graph)
        is_primary_node_selected = \
            distance_to_root_level(graph, node_from_graph) >= requirements.max_depth and randint(0, 1)
    if is_primary_node_selected:
        new_subtree = graph_gen_params.node_factory.get_node(is_primary=True)
        if not new_subtree:
            return graph
    else:
        new_subtree = graph_gen_params.random_graph_factory(requirements, max_depth).root_node
    graph.update_subtree(node_from_graph, new_subtree)
    return graph


@register_native
def growth_mutation(graph: OptGraph,
                    requirements: GraphRequirements,
                    graph_gen_params: GraphGenerationParams,
                    parameters: AlgorithmParameters,
                    local_growth: bool = True
                    ) -> OptGraph:
    """
    This mutation adds new nodes to the graph (just single node between existing nodes or new subtree).

    :param graph: graph to mutate
    :param local_growth: if true then maximal depth of new subtree equals depth of tree located in
    selected random node, if false then previous depth of selected node doesn't affect to
    new subtree depth, maximal depth of new subtree just should satisfy depth constraint in parent tree
    """

    if random() > 0.5:
        # simple growth (one node can be added)
        return single_add_mutation(graph, requirements, graph_gen_params, parameters)
    else:
        # advanced growth (several nodes can be added)
        return tree_growth(graph, requirements, graph_gen_params, parameters, local_growth)


@register_native
def reduce_mutation(graph: OptGraph,
                    requirements: GraphRequirements,
                    graph_gen_params: GraphGenerationParams,
                    parameters: AlgorithmParameters,
                    ) -> OptGraph:
    """
    Selects a random node in a tree, then removes its subtree. If the current arity of the node's
    parent is more than the specified minimal arity, then the selected node is also removed.
    Otherwise, it is replaced by a random primary node.

    :param graph: graph to mutate
    """
    if len(graph.nodes) == 1:
        return graph

    nodes = [node for node in graph.nodes if node is not graph.root_node]
    node_to_del = choice(nodes)
    children = graph.node_children(node_to_del)
    is_possible_to_delete = all([len(child.nodes_from) - 1 >= requirements.min_arity for child in children])
    if is_possible_to_delete:
        graph.delete_subtree(node_to_del)
    else:
        primary_node = graph_gen_params.node_factory.get_node(is_primary=True)
        if not primary_node:
            return graph
        graph.update_subtree(node_to_del, primary_node)
    return graph


@register_native
def no_mutation(graph: OptGraph, *args, **kwargs) -> OptGraph:
    return graph


base_mutations_repo = {
    MutationTypesEnum.none: no_mutation,
    MutationTypesEnum.simple: simple_mutation,
    MutationTypesEnum.growth: partial(growth_mutation, local_growth=False),
    MutationTypesEnum.local_growth: partial(growth_mutation, local_growth=True),
    MutationTypesEnum.tree_growth: tree_growth,
    MutationTypesEnum.reduce: reduce_mutation,
    MutationTypesEnum.single_add: single_add_mutation,
    MutationTypesEnum.single_edge: single_edge_mutation,
    MutationTypesEnum.single_drop: single_drop_mutation,
    MutationTypesEnum.single_change: single_change_mutation
}

simple_mutation_set = (
    MutationTypesEnum.tree_growth,
    MutationTypesEnum.single_add,
    MutationTypesEnum.single_change,
    MutationTypesEnum.single_drop,
    MutationTypesEnum.single_edge,
    # join nodes
    # flip edge
    # cycle edge
)

rich_mutation_set = (
    MutationTypesEnum.simple,
    MutationTypesEnum.reduce,
    MutationTypesEnum.growth,
    MutationTypesEnum.local_growth
)



import io

import networkx as nx
from PIL import Image
from rdkit import Chem
from rdkit.Chem import MolFromSmiles, MolToSmiles, SanitizeMol, Kekulize, MolToInchi
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.rdchem import Atom, BondType, RWMol, GetPeriodicTable, ChiralType, HybridizationType


class MolGraph:
    def __init__(self, rw_molecule: RWMol):
        self._rw_molecule = rw_molecule
        self.update_representation()

    @staticmethod
    def from_smiles(smiles: str):
        molecule = MolFromSmiles(smiles)
        rw_molecule = RWMol(molecule)
        return MolGraph(rw_molecule)

    @staticmethod
    def from_nx_graph(graph: nx.Graph):
        """Original code: https://github.com/maxhodak/keras-molecules"""
        mol = RWMol()
        atomic_nums = nx.get_node_attributes(graph, 'atomic_num')
        chiral_tags = nx.get_node_attributes(graph, 'chiral_tag')
        formal_charges = nx.get_node_attributes(graph, 'formal_charge')
        node_is_aromatics = nx.get_node_attributes(graph, 'is_aromatic')
        node_hybridizations = nx.get_node_attributes(graph, 'hybridization')
        num_explicit_hss = nx.get_node_attributes(graph, 'num_explicit_hs')
        node_to_idx = {}
        for node in graph.nodes():
            a = Chem.Atom(atomic_nums[node])
            a.SetChiralTag(ChiralType(chiral_tags[node]))
            a.SetFormalCharge(formal_charges[node])
            a.SetIsAromatic(node_is_aromatics[node])
            a.SetHybridization(HybridizationType(node_hybridizations[node]))
            a.SetNumExplicitHs(num_explicit_hss[node])
            idx = mol.AddAtom(a)
            node_to_idx[node] = idx

        bond_types = nx.get_edge_attributes(graph, 'bond_type')
        for edge in graph.edges():
            first, second = edge
            ifirst = node_to_idx[first]
            isecond = node_to_idx[second]
            bond_type = BondType(bond_types[first, second])
            mol.AddBond(ifirst, isecond, bond_type)

        SanitizeMol(mol)
        return MolGraph(mol)

    @property
    def heavy_atoms_number(self) -> int:
        return self._rw_molecule.GetNumAtoms()

    def get_nx_graph(self) -> nx.Graph:
        """Original code: https://github.com/maxhodak/keras-molecules"""
        graph = nx.Graph()

        for atom in self._rw_molecule.GetAtoms():
            graph.add_node(atom.GetIdx(),
                           # save indices in node data to restore edges parameters after adapt-restore process
                           nxid=str(atom.GetIdx()),
                           name=atom.GetSymbol(),
                           atomic_num=atom.GetAtomicNum(),
                           formal_charge=atom.GetFormalCharge(),
                           chiral_tag=atom.GetChiralTag(),
                           hybridization=atom.GetHybridization(),
                           num_explicit_hs=atom.GetNumExplicitHs(),
                           is_aromatic=atom.GetIsAromatic())
        for bond in self._rw_molecule.GetBonds():
            graph.add_edge(bond.GetBeginAtomIdx(),
                           bond.GetEndAtomIdx(),
                           bond_type=bond.GetBondType())
        return graph

    def get_smiles(self, aromatic: bool = False) -> str:
        return MolToSmiles(self.get_rw_molecule(aromatic))

    def get_rw_molecule(self, aromatic: bool = False) -> RWMol:
        if aromatic:
            return MolFromSmiles(self.get_smiles())
        else:
            return self._rw_molecule

    def update_representation(self):
        SanitizeMol(self._rw_molecule)
        Kekulize(self._rw_molecule)

        # Setting all bonds and atoms to non aromatics
        for atom_from in self._rw_molecule.GetAtoms():
            atom_from.SetIsAromatic(False)
            for atom_to in self._rw_molecule.GetAtoms():
                bond = self._rw_molecule.GetBondBetweenAtoms(atom_from.GetIdx(), atom_to.GetIdx())
                if bond is not None:
                    bond.SetIsAromatic(False)

        # Updating the property cache of atoms
        for atom in self._rw_molecule.GetAtoms():
            atom.UpdatePropertyCache()

        self._rw_molecule.UpdatePropertyCache()

    def add_atom(self, atom_type: str):
        atom = Atom(atom_type)
        self._rw_molecule.AddAtom(atom)

    def set_bond(self, from_atom: int, to_atom: int, bond_type: BondType = BondType.SINGLE, update_representation=True):
        current_bond = self._rw_molecule.GetBondBetweenAtoms(from_atom, to_atom)
        if current_bond is None:
            self._rw_molecule.AddBond(from_atom, to_atom, bond_type)
        else:
            current_bond.SetBondType(bond_type)
        if update_representation:
            self.update_representation()

    def remove_bond(self, from_atom: int, to_atom: int, update_representation=True):
        self._rw_molecule.RemoveBond(from_atom, to_atom)
        if update_representation:
            self.update_representation()

    def remove_atom(self, atom_id: int, update_representation=True):
        self._rw_molecule.RemoveAtom(atom_id)
        if update_representation:
            self.update_representation()

    def replace_atom(self, atom_to_replace: int, atom_type: str):
        new_atomic_number = GetPeriodicTable().GetAtomicNumber(atom_type)

        self._rw_molecule.GetAtomWithIdx(atom_to_replace).SetAtomicNum(new_atomic_number)

        self._rw_molecule.GetAtomWithIdx(atom_to_replace).SetFormalCharge(0)

        self.update_representation()

    def show(self):
        drawer = rdMolDraw2D.MolDraw2DCairo(300, 300)
        drawer.drawOptions().addStereoAnnotation = False
        drawer.DrawMolecule(self._rw_molecule)
        drawer.FinishDrawing()
        bytes_images = drawer.GetDrawingText()
        image = Image.open(io.BytesIO(bytes_images))
        image.show()

    def __eq__(self, other_mol: 'MolGraph'):
        return MolToInchi(self.get_rw_molecule(aromatic=True)) == MolToInchi(other_mol.get_rw_molecule(aromatic=True))

    def __hash__(self):
        return hash(MolToInchi(self.get_rw_molecule(aromatic=True)))

    def __str__(self):
        return self.get_smiles()



#    This code is modified part of DEAP library (Library URL: https://github.com/DEAP/deap).
import sys
from numbers import Real
from typing import Sequence, Union
from operator import mul, truediv

import numpy as np

from golem.core.optimisers.fitness.fitness import Fitness, is_metric_worse


class MultiObjFitness(Fitness):
    """The fitness is a measure of quality of a solution. If *values* are
    provided as a tuple, the fitness is initialized using those values,
    otherwise it is empty (or invalid).

    :param values: The initial values of the fitness as a tuple, optional.
    :param weights: Optional weights for fitness values. 1.0 by default.
    :param wvalues: Allows to provide values by giving already weighted values.
    If provided, then :param values: is ignored. This is used by deserialization.

    Comparison between Fitnesses is made lexicographically.
    Maximization and minimization are taken care off by a multiplication
    between the :attr:`weights` and the fitness :attr:`values`. The comparison
    can be made between fitnesses of different size, if the fitnesses are
    equal until the extra elements, the longer fitness will be superior to the
    shorter.
    """

    def __init__(self, values: Sequence[Real] = (),
                 weights: Union[Sequence[Real], Real] = 1,
                 *, wvalues: Sequence[Real] = None):
        if wvalues is not None:
            # This branch is mainly for deserialization from .wvalues attribute
            values = tuple(np.true_divide(wvalues, weights))

        if isinstance(weights, Real):
            # Single value provided or default 1.0 weights
            weights = (weights,) * len(values)
        elif isinstance(weights, Sequence):
            self._check_length(values, weights)
        else:
            raise TypeError("Attribute weights of %r must be a sequence or a number." % self.__class__)

        self._weights = weights
        self.values = values

    @property
    def weights(self) -> Sequence[Real]:
        return self._weights

    def getValues(self):
        return self.wvalues

    def setValues(self, values):
        if values is None:
            self.reset()
        else:
            self._check_length(values)
        try:
            self.wvalues = tuple(map(mul, values, self.weights))
        except TypeError:
            _, _, traceback = sys.exc_info()
            raise TypeError("Both weights and assigned values must be a "
                            "sequence of numbers when assigning to values of "
                            "%r. Currently assigning value(s) %r of %r to a "
                            "fitness with weights %s."
                            % (self.__class__, values, type(values),
                               self.weights)).with_traceback(traceback)

    def delValues(self):
        self.wvalues = ()

    values = property(getValues, setValues, delValues,
                      ("Fitness values. Use directly ``individual.fitness.values = values`` "
                       "in order to set the fitness and ``del individual.fitness.values`` "
                       "in order to clear (invalidate) the fitness."))

    def dominates(self, other: 'MultiObjFitness', selector=slice(None)):
        """Return true if each objective of *self* is not strictly worse than
        the corresponding objective of *other* and at least one objective is
        strictly better.

        :param other: Other multi-objective fitness for comparison
        :param selector: Slice indicating on which objectives the domination is
                    tested. The default value is `slice(None)`, representing
                    every objectives.
        """
        not_equal = False
        for self_wvalue, other_wvalue in zip(self.wvalues[selector], other.wvalues[selector]):
            if is_metric_worse(other_wvalue, self_wvalue):
                not_equal = True
            elif is_metric_worse(self_wvalue, other_wvalue):
                return False
        return not_equal

    @property
    def valid(self):
        """Assess if a fitness is valid or not."""
        return len(self.wvalues) != 0

    def __hash__(self):
        return hash(self.wvalues)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.valid and other.valid and
                self.allclose(self.wvalues, other.wvalues))

    def __str__(self):
        return str(tuple(round(wval, 4) for wval in self.wvalues))

    def _check_length(self, values, weights=None):
        if weights is None:
            weights = self.weights
        if len(values) != len(weights):
            raise TypeError("Attribute weights for all values must be provided.")



from golem.core.optimisers.graph import OptNode, OptGraph


def get_opt_graph():
    node = OptNode('1')
    node2 = OptNode('2')
    node4 = OptNode('4', nodes_from=[node, node2])
    node3 = OptNode('3')
    node5 = OptNode('5', nodes_from=[node4, node3])
    graph = OptGraph(node5)
    return graph


if __name__ == "__main__":
    # Shows how to set labels for nodes and edges.
    graph = get_opt_graph()

    nodes_labels = {}
    for i, node in enumerate(graph.nodes):
        nodes_labels[i] = f'{node.name}_node'

    edges_labels = {}
    for i, edge in enumerate(graph.get_edges()):
        edges_labels[i] = f'{edge[0].name}_{edge[1].name}'

    graph.show(nodes_labels=nodes_labels, edges_labels=edges_labels)



from enum import Enum

from typing import List, Callable, Any

import numpy as np

from golem.utilities.requirements_notificator import warn_requirement
from golem.core.adapter.nx_adapter import BanditNetworkxAdapter
from golem.core.optimisers.opt_history_objects.individual import Individual

try:
    from karateclub import FeatherGraph
except ModuleNotFoundError:
    warn_requirement('karateclub', 'other_requirements/adaptive.txt')


def adapter_func_to_networkx(func):
    """ Decorator function to adapt observation to networkx graphs. """
    def wrapper(obs, available_operations):
        nx_graph = BanditNetworkxAdapter().restore(obs)
        embedding = func(nx_graph, available_operations)
        return embedding
    return wrapper


def adapter_func_to_graph(func):
    """ Decorator function to adapt observation to networkx graphs. """
    def wrapper(obs, available_operations):
        if isinstance(obs, Individual):
            graph = obs.graph
        else:
            graph = obs
        return func(graph, available_operations)
    return wrapper


def encode_operations(operations: List[str], available_operations: List[str], mode: str = 'label'):
    """ Encoding of operations.
    :param operations: operations to encode
    :param available_operations: list of all available operations
    :param mode: mode of encoding. Available type: 'OHE' and 'label', default -- 'label'
    """
    encoded = []
    for operation in operations:
        if mode == 'label':
            encoding = available_operations.index(operation)
        else:
            encoding = [0] * len(available_operations)
            encoding[available_operations.index(operation)] = 1
        encoded.append(encoding)
    return encoded


@adapter_func_to_networkx
def feather_graph(obs: Any, available_operations: List[str]) -> List[float]:
    """ Returns embedding based on an implementation of `"FEATHER-G" <https://arxiv.org/abs/2005.07959>`_.
    The procedure uses characteristic functions of node features with random walk weights to describe
    node neighborhoods. These node level features are pooled by mean pooling to
    create graph level statistics. """
    descriptor = FeatherGraph()
    descriptor.fit([obs])
    emb = descriptor.get_embedding().reshape(-1, 1)
    embd = [i[0] for i in emb]
    return embd


@adapter_func_to_graph
def nodes_num(obs: Any, available_operations: List[str]) -> List[int]:
    """ Returns number of nodes in graph. """
    return [len(obs.nodes)]


@adapter_func_to_graph
def labeled_edges(obs: Any, available_operations: List[str]) -> List[int]:
    """ Encodes graph with its edges with nodes labels. """
    operations = []
    for node in obs.nodes:
        for node_ in node.nodes_from:
            operations.append(node_.name)
            operations.append(node.name)
    return encode_operations(operations=operations, available_operations=available_operations)


@adapter_func_to_graph
def operations_quantity(obs: Any, available_operations: List[str]) -> List[int]:
    """ Encodes graphs as vectors with quantity of each operation. """
    encoding = [0] * len(available_operations)
    for node in obs.nodes:
        encoding[available_operations.index(node.name)] += 1
    return encoding


@adapter_func_to_graph
def adjacency_matrix(obs: Any, available_operations: List[str]) -> List[int]:
    """ Encodes graphs as flattened adjacency matrix. """
    matrix = np.zeros((len(available_operations), len(available_operations)))
    for node in obs.nodes:
        operation_parent_idx = available_operations.index(node.name)
        for node_ in node.nodes_from:
            operation_child_idx = available_operations.index(node_.name)
            matrix[operation_parent_idx][operation_child_idx] += 1
    return matrix.reshape(1, -1)[0].astype(int).tolist()


def none_encoding(obs: Any, available_operations: List[str]) -> List[int]:
    """ Empty encoding. """
    return obs


class ContextAgentTypeEnum(Enum):
    feather_graph = 'feather_graph'
    nodes_num = 'nodes_num'
    labeled_edges = 'labeled_edges'
    operations_quantity = 'operations_quantity'
    adjacency_matrix = 'adjacency_matrix'
    none_encoding = 'none_encoding'


class ContextAgentsRepository:
    """ Repository of functions to encode observations. """
    _agents_implementations = {
        ContextAgentTypeEnum.feather_graph: feather_graph,
        ContextAgentTypeEnum.nodes_num: nodes_num,
        ContextAgentTypeEnum.labeled_edges: labeled_edges,
        ContextAgentTypeEnum.operations_quantity: operations_quantity,
        ContextAgentTypeEnum.adjacency_matrix: adjacency_matrix,
        ContextAgentTypeEnum.none_encoding: none_encoding
    }

    @staticmethod
    def agent_class_by_id(agent_id: ContextAgentTypeEnum) -> Callable:
        return ContextAgentsRepository._agents_implementations[agent_id]



from typing import Sequence, List, TYPE_CHECKING, Callable, Union

from golem.utilities.data_structures import ensure_wrapped_in_sequence

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.dag.graph_node import GraphNode


def distance_to_root_level(graph: 'Graph', node: 'GraphNode') -> int:
    """Gets distance to the final output node

    Args:
        graph: graph for finding the distance
        node: search starting point

    Return:
        int: distance to root level
    """

    def child_height(parent_node: 'GraphNode') -> int:
        height = 0
        for _ in range(graph.length):
            node_children = graph.node_children(parent_node)
            if node_children:
                height += 1
                parent_node = node_children[0]
            else:
                return height

    if graph_has_cycle(graph):
        return -1
    height = child_height(node)
    return height


def distance_to_primary_level(node: 'GraphNode') -> int:
    depth = node_depth(node)
    return depth - 1 if depth > 0 else -1


def nodes_from_layer(graph: 'Graph', layer_number: int) -> Sequence['GraphNode']:
    """Gets all the nodes from the chosen layer up to the surface

    Args:
        graph: graph with nodes
        layer_number: max height of diving

    Returns:
        all nodes from the surface to the ``layer_number`` layer
    """

    def get_nodes(roots: Sequence['GraphNode'], current_height: int) -> Sequence['GraphNode']:
        """Gets all the parent nodes of ``roots``

        :param roots: nodes to get all subnodes from
        :param current_height: current diving step depth

        :return: all parent nodes of ``roots`` in one sequence:69
        """
        nodes = []
        if current_height == layer_number:
            nodes.extend(roots)
        else:
            for root in roots:
                nodes.extend(get_nodes(root.nodes_from, current_height + 1))
        return nodes

    nodes = get_nodes(graph.root_nodes(), current_height=0)
    return nodes


def ordered_subnodes_hierarchy(node: 'GraphNode') -> List['GraphNode']:
    """Gets hierarchical subnodes representation of the graph starting from the bounded node

    Returns:
        List['GraphNode']: hierarchical subnodes list starting from the bounded node
    """
    started = {node}
    visited = set()

    def subtree_impl(node):
        nodes = [node]
        for parent in node.nodes_from:
            if parent in visited:
                continue
            elif parent in started:
                raise ValueError('Can not build ordered node hierarchy: graph has cycle')
            started.add(parent)
            nodes.extend(subtree_impl(parent))
            visited.add(parent)
        return nodes

    return subtree_impl(node)


def node_depth(nodes: Union['GraphNode', Sequence['GraphNode']]) -> int:
    """Gets the maximal depth among the provided ``nodes`` in the graph

    Args:
        nodes: nodes to calculate the depth for

    Returns:
        int: maximal depth
    """
    nodes = ensure_wrapped_in_sequence(nodes)
    final_depth = {}
    subnodes = set()
    for node in nodes:
        max_depth = 0
        # if node is a subnode of another node it has smaller depth
        if node.uid in subnodes:
            continue
        depth = 1
        visited = []
        if node in visited:
            return -1
        visited.append(node)
        stack = [(node, depth, iter(node.nodes_from))]
        while stack:
            curr_node, depth_now, parents = stack[-1]
            try:
                parent = next(parents)
                subnodes.add(parent.uid)
                if parent not in visited:
                    visited.append(parent)
                    if parent.uid in final_depth:
                        # depth of the parent has been already calculated
                        stack.append((parent, depth_now + final_depth[parent.uid], iter([])))
                    else:
                        stack.append((parent, depth_now + 1, iter(parent.nodes_from)))
                else:
                    return -1
            except StopIteration:
                _, depth_now, _ = stack.pop()
                visited.pop()
                max_depth = max(max_depth, depth_now)
        final_depth[node.uid] = max_depth
    return max(final_depth.values())


def map_dag_nodes(transform: Callable, nodes: Sequence) -> Sequence:
    """Maps nodes in dfs-order while respecting node edges.

    Args:
        transform: node transform function (maps node to node)
        nodes: sequence of nodes for mapping

    Returns:
        Sequence: sequence of transformed links with preserved relations
    """
    mapped_nodes = {}

    def map_impl(node):
        already_mapped = mapped_nodes.get(id(node))
        if already_mapped:
            return already_mapped
        # map node itself
        mapped_node = transform(node)
        # remember it to avoid recursion
        mapped_nodes[id(node)] = mapped_node
        # map its children
        mapped_node.nodes_from = list(map(map_impl, node.nodes_from))
        return mapped_node

    return list(map(map_impl, nodes))


def graph_structure(graph: 'Graph') -> str:
    """ Returns structural information about the graph - names and parameters of graph nodes.
    Represents graph info in easily readable way.

    Returns:
        str: graph structure
    """
    return '\n'.join([str(graph), *(f'{node.name} - {node.parameters}' for node in graph.nodes)])


def graph_has_cycle(graph: 'Graph') -> bool:
    """ Returns True if the graph contains a cycle and False otherwise. Implements Depth-First Search."""

    visited = {node.uid: False for node in graph.nodes}
    stack = []
    on_stack = {node.uid: False for node in graph.nodes}
    for node in graph.nodes:
        if visited[node.uid]:
            continue
        stack.append(node)
        while len(stack) > 0:
            cur_node = stack[-1]
            if not visited[cur_node.uid]:
                visited[cur_node.uid] = True
                on_stack[cur_node.uid] = True
            else:
                on_stack[cur_node.uid] = False
                stack.pop()
            for parent in cur_node.nodes_from:
                if not visited[parent.uid]:
                    stack.append(parent)
                elif on_stack[parent.uid]:
                    return True
    return False



import random
from typing import Optional

import numpy as np

from golem.core.log import default_log


class RandomStateHandler:
    MODEL_FITTING_SEED = 0

    @staticmethod
    def log_random_state():
        py_state = random.getstate()
        np_state = np.random.get_state()
        log = default_log(RandomStateHandler.__class__.__name__)
        log.debug(f'Random State: random.getstate() follows...\n{py_state}')
        log.debug(f'Random State: numpy.random.get_state() follows...\n{np_state}')

    def __init__(self, seed: Optional[int] = None):
        if seed is None:
            seed = RandomStateHandler.MODEL_FITTING_SEED
        self._seed = seed
        self._old_seed = None

    def __enter__(self):
        self._old_np_state = np.random.get_state()
        self._old_state = random.getstate()

        np.random.seed(self._seed)
        random.seed(self._seed)
        return self._seed

    def __exit__(self, exc_type, exc_value, exc_traceback):
        np.random.set_state(self._old_np_state)
        random.setstate(self._old_state)



from __future__ import annotations

from collections import UserList
from copy import deepcopy, copy
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Union

from golem.utilities.data_structures import ensure_wrapped_in_sequence

if TYPE_CHECKING:
    from golem.core.optimisers.opt_history_objects.individual import Individual


class Generation(UserList):
    """
    List of evolution individuals considered as a single generation.
    Allows to provide additional information about the generation.
    Responsible for setting generation-related info to the individuals.
    """

    def __init__(self, iterable: Union[Iterable[Individual], Generation], generation_num: int,
                 label: Optional[str] = None, metadata: Optional[Dict[str, Any]] = None):
        self.generation_num = generation_num
        self.label: str = label or ''
        self.metadata: Dict[str, Any] = metadata or {}
        super().__init__(iterable)
        self._set_native_generation(self.data)

    def __setitem__(self, index, item: Union[Individual, Iterable[Individual]]):
        super().__setitem__(index, item)
        self._set_native_generation(item)

    def copy(self) -> Generation:
        return copy(self)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.data = copy(self.data)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            object.__setattr__(result, k, deepcopy(v, memo))
        return result

    def _set_native_generation(self, individuals: Union[Individual, Iterable[Individual]]):
        individuals = ensure_wrapped_in_sequence(individuals)
        for individual in individuals:
            individual.set_native_generation(self.generation_num)

    def __repr__(self):
        gen_num = f'Generation {self.generation_num}'
        label = f' ({self.label}): ' if self.label else ': '
        data = super().__repr__()
        return ''.join([gen_num, label, data])



import json
import logging
import multiprocessing
import os
import pathlib
import sys
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler
from typing import Optional, Tuple, Union

from golem.utilities.singleton_meta import SingletonMeta
from golem.core.paths import default_data_dir

DEFAULT_LOG_PATH = pathlib.Path(default_data_dir(), 'log.log')


class Log(metaclass=SingletonMeta):
    """Log object to store logger singleton and log adapters

    Args:
        config_json_file: ``json`` file from which to collect the logger if specified
        output_logging_level: logging levels are the same as in standard python module 'logging'
        log_file: file to write logs in
    """

    __log_adapters = {}

    def __init__(self,
                 config_json_file: str = 'default',
                 output_logging_level: int = logging.INFO,
                 log_file: Optional[Union[str, pathlib.Path]] = None,
                 use_console: bool = True):
        self.log_file = log_file or DEFAULT_LOG_PATH
        self.logger = self._get_logger(config_file=config_json_file,
                                       logging_level=output_logging_level,
                                       use_console=use_console)

    @staticmethod
    def setup_in_mp(logging_level: int, logs_dir: pathlib.Path):
        """
        Preserves logger level and its records in a separate file for each process only if it's a child one

        Args:
            logging_level: level of the logger from the main process
            logs_dir: path to the logs directory
        """

        cur_proc = multiprocessing.current_process().name
        log_file_name = logs_dir.joinpath(f'log_{cur_proc}.log')
        Log(output_logging_level=logging_level, log_file=log_file_name, use_console=False)

    def get_parameters(self) -> Tuple[int, pathlib.Path]:
        return self.logger.level, pathlib.Path(self.log_file).parent

    def reset_logging_level(self, logging_level: int):
        """ Resets logging level for logger and its handlers """
        # Resets logging level is needed because before initialization with API params Singleton
        # can be initialized somewhere else with default ones
        self.logger.setLevel(logging_level)
        for handler in self.handlers:
            handler.setLevel(logging_level)
        for adapter in self.__log_adapters.values():
            adapter.logging_level = logging_level

    def get_adapter(self, prefix: str) -> 'LoggerAdapter':
        """ Get adapter to pass contextual information to log messages

        Args:
            prefix: prefix to log messages with this adapter. Usually, the prefix is the name of the class
                where the log came from
        """

        if prefix not in self.__log_adapters.keys():
            self.__log_adapters[prefix] = LoggerAdapter(self.logger,
                                                        {'prefix': prefix})
        return self.__log_adapters[prefix]

    def _get_logger(self, config_file: str, logging_level: int, use_console: bool = True) -> logging.Logger:
        """ Get logger object """
        logger = logging.getLogger()
        if config_file != 'default':
            self._setup_logger_from_json_file(config_file)
        else:
            logger = self._setup_default_logger(logger=logger, logging_level=logging_level, use_console=use_console)
        return logger

    def _setup_default_logger(self, logger: logging.Logger, logging_level: int,
                              use_console: bool = True) -> logging.Logger:
        """ Define console and file handlers for logger
        """

        if use_console:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setLevel(logging_level)
            console_formatter = logging.Formatter('%(asctime)s - %(message)s')
            console_handler.setFormatter(console_formatter)
            logger.addHandler(console_handler)

        file_handler = RotatingFileHandler(self.log_file, maxBytes=100000000, backupCount=1)
        file_handler.setLevel(logging_level)
        file_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logger.addHandler(file_handler)

        logger.setLevel(logging_level)

        return logger

    @staticmethod
    def _setup_logger_from_json_file(config_file):
        """ Setup logging configuration from file
        """

        try:
            with open(config_file, 'rt') as file:
                config = json.load(file)
            dictConfig(config)
        except Exception as ex:
            raise Exception(f'Can not open the log config file because of {ex}')

    @property
    def handlers(self):
        return self.logger.handlers

    def release_handlers(self):
        """ This function closes handlers of logger
        """

        for handler in self.handlers:
            handler.close()

    def getstate(self):
        """ Define the attributes to be pickled via deepcopy or pickle

        Returns:
            dict: ``dict`` of state
        """

        state = dict(self.__dict__)
        del state['logger']
        return state

    def __str__(self):
        return f'Log object for {self.logger.name} module'

    def __repr__(self):
        return self.__str__()


class LoggerAdapter(logging.LoggerAdapter):
    """ This class looks like logger but used to pass contextual information
    to the output along with logging event information
    """

    def __init__(self, logger: logging.Logger, extra: dict):
        super().__init__(logger=logger, extra=extra)
        self.logging_level = logger.level
        self.setLevel(self.logging_level)

    def process(self, msg, kwargs):
        self.logger.setLevel(self.logging_level)
        return '%s - %s' % (self.extra['prefix'], msg), kwargs

    def debug(self, msg, *args, **kwargs):
        raise_if_test(msg, **kwargs)
        super().debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        raise_if_test(msg, **kwargs)
        super().info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        raise_if_test(msg, **kwargs)
        super().warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        raise_if_test(msg, **kwargs)
        super().error(msg, *args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        raise_if_test(msg, **kwargs)
        super().exception(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        raise_if_test(msg, **kwargs)
        super().critical(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """
        Delegate a log call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        raise_if_test(msg, **kwargs)
        super().log(level, msg, *args, **kwargs)

    def message(self, msg: str, **kwargs):
        """ Record the message to user.
        Message is an intermediate logging level between info and warning
        to display main info about optimization process """
        raise_if_test(msg, **kwargs)
        message_logging_level = 45
        if message_logging_level >= self.logging_level:
            self.critical(msg=msg)

    def __str__(self):
        return f'LoggerAdapter object for {self.extra["prefix"]} module'

    def __repr__(self):
        return self.__str__()


def raise_if_test(msg, **kwargs):
    if kwargs.get('raise_if_test', False) is True and is_test_session:
        raise Exception(msg)


def is_test_session():
    return 'PYTEST_CURRENT_TEST' in os.environ


def default_log(prefix: Optional[object] = 'default') -> 'LoggerAdapter':
    """ Default logger

    Args:
        prefix: adapter prefix to add it to log messages

    Returns:
        :obj:`LoggerAdapter`: :obj:`LoggerAdapter` object

    """

    # get log prefix
    if not isinstance(prefix, str):
        prefix = prefix.__class__.__name__

    return Log().get_adapter(prefix=prefix)



from os.path import join
from typing import Optional, List, Type

import multiprocessing

from golem.core.log import default_log
from golem.core.dag.graph import Graph
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.timer import OptimisationTimer
from golem.core.paths import default_data_dir
from golem.structural_analysis.graph_sa.edge_sa_approaches import EdgeAnalyzeApproach, EdgeAnalysis
from golem.structural_analysis.graph_sa.entities.edge import Edge
from golem.structural_analysis.graph_sa.results.sa_analysis_results import SAAnalysisResults
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements


class EdgesAnalysis:
    """
    This class is for edges structural analysis within an Graph .
    It takes edges and approaches to be applied to chosen edges.
    To define which edges to analyze pass them to edges_to_analyze filed
    or all edges will be analyzed.

    :param objective: list of objective functions for computing metric values
    :param approaches: methods applied to edges to modify the graph or analyze certain operations.\
    Default: [EdgeDeletionAnalyze, EdgeReplaceOperationAnalyze]
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural
    """

    def __init__(self, objective: Objective,
                 approaches: Optional[List[Type[EdgeAnalyzeApproach]]] = None,
                 requirements: Optional[StructuralAnalysisRequirements] = None,
                 path_to_save: Optional[str] = None):

        self.objective = objective
        self.approaches = approaches
        self.requirements = \
            StructuralAnalysisRequirements() if requirements is None else requirements
        self.log = default_log(self)
        self.path_to_save = \
            join(default_data_dir(), 'structural', 'edges_structural') if path_to_save is None else path_to_save

    def analyze(self, graph: Graph, results: Optional[SAAnalysisResults] = None,
                edges_to_analyze: Optional[List[Edge]] = None,
                n_jobs: int = 1, timer: Optional[OptimisationTimer] = None) -> SAAnalysisResults:
        """
        Main method to run the analyze process for every edge.

        :param graph: graph object to analyze
        :param results: SA results
        :param edges_to_analyze: edges to analyze. Default: all edges
        :param n_jobs: n_jobs
        :param timer: timer indicating how much time is left for optimization
        :return edges_results: dict with analysis result per Edge
        """

        if not results:
            results = SAAnalysisResults()

        if n_jobs == -1:
            n_jobs = multiprocessing.cpu_count()

        if not edges_to_analyze:
            self.log.message('Edges to analyze are not defined. All edges will be analyzed.')
            edges_to_analyze = [Edge.from_tuple([edge])[0] for edge in graph.get_edges()]

        edge_analysis = EdgeAnalysis(approaches=self.approaches,
                                     approaches_requirements=self.requirements,
                                     path_to_save=self.path_to_save)

        with multiprocessing.Pool(processes=n_jobs) as pool:
            cur_edges_result = pool.starmap(edge_analysis.analyze,
                                            [[graph, edge, self.objective, timer]
                                             for edge in edges_to_analyze])
        results.add_results(cur_edges_result)

        return results



import numpy as np


def edit_distance(A1, A2):
    """The edit distance between graphs, defined as the number of changes one
    needs to make to put the edge lists in correspondence.

    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared

    Returns
    -------
    dist : float
        The edit distance between the two graphs
    """
    dist = np.abs((A1 - A2)).sum() / 2
    return dist



from dataclasses import dataclass
from typing import Sequence, Union, Any, Optional, Callable

from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum
from golem.core.optimisers.adaptive.mab_agents.neural_contextual_mab_agent import ContextAgentTypeEnum
from golem.core.optimisers.genetic.operators.base_mutations import MutationStrengthEnum, MutationTypesEnum, \
    simple_mutation_set
from golem.core.optimisers.optimizer import AlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.elitism import ElitismTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.regularization import RegularizationTypesEnum
from golem.core.optimisers.genetic.operators.selection import SelectionTypesEnum


@dataclass
class GPAlgorithmParameters(AlgorithmParameters):
    """
    Defines parameters of evolutionary operators and the algorithm of genetic optimizer.

    :param crossover_prob: crossover probability (chance that two individuals will be mated).
    :param mutation_prob: mutation probability (chance that an individual will be mutated).
    :param variable_mutation_num: flag to apply mutation one or few times for individual in each iteration.
    :param max_num_of_operator_attempts: max number of unsuccessful evo operator attempts before continuing.
    :param mutation_strength: strength of mutation in tree (using in certain mutation types)
    :param min_pop_size_with_elitism: minimal population size with which elitism is applicable
    :param required_valid_ratio: ratio of valid individuals on next population to continue optimization.

    Used in `ReproductionController` to compensate for invalid individuals. See the class for details.

    :param adaptive_mutation_type: Experimental feature! Enables adaptive Mutation agent.
    :param context_agent_type: Experimental feature! Enables graph encoding for Mutation agent.

    Adaptive mutation agent uses specified algorithm. 'random' type is the default non-adaptive version.
    Requires crossover_types to be CrossoverTypesEnum.none for correct adaptive learning,
    so that fitness changes depend only on agent's actions (chosen mutations).
    ``MutationAgentTypeEnum.bandit`` uses Multi-Armed Bandit (MAB) learning algorithm.
    ``MutationAgentTypeEnum.contextual_bandit`` uses contextual MAB learning algorithm.
    ``MutationAgentTypeEnum.neural_bandit`` uses contextual MAB learning algorithm with Deep Neural encoding.

    Parameter `context_agent_type` specifies implementation of graph/node encoder for adaptive
    mutation agent. It is relevant for contextual and neural bandits.

    :param selection_types: Sequence of selection operators types
    :param crossover_types: Sequence of crossover operators types
    :param mutation_types: Sequence of mutation operators types
    :param elitism_type: type of elitism operator evolution

    :param regularization_type: type of regularization operator

    Regularization attempts to cut off the subtrees of the graph. If the truncated graph
    is not worse than the original, then it enters the new generation as a simpler solution.
    Regularization is not used by default, it must be explicitly enabled.

    :param genetic_scheme_type: type of genetic evolutionary scheme

    The `generational` scheme is a standard scheme of the evolutionary algorithm.
    It specifies that at each iteration the entire generation is updated.

    In the `steady_state` scheme at each iteration only one individual is updated.

    The `parameter_free` scheme is an adaptive variation of the `steady_state` scheme.
    It specifies that the population size and the probability of mutation and crossover
    change depending on the success of convergence. If there are no improvements in fitness,
    then the size and the probabilities increase. When fitness improves, the size and the
    probabilities decrease. That is, the algorithm choose a more stable and conservative
    mode when optimization seems to converge.

    :param decaying_factor: decaying factor for Multi-Armed Bandits for managing the profit from operators
        The smaller the value of decaying_factor, the larger the influence for the best operator.
    :param window_size: the size of sliding window for Multi-Armed Bandits to decrease variance.
        The window size is measured by the number of individuals to consider.
    """

    crossover_prob: float = 0.8
    mutation_prob: float = 0.8
    variable_mutation_num: bool = True
    max_num_of_operator_attempts: int = 100
    mutation_strength: MutationStrengthEnum = MutationStrengthEnum.mean
    min_pop_size_with_elitism: int = 5
    required_valid_ratio: float = 0.9

    adaptive_mutation_type: MutationAgentTypeEnum = MutationAgentTypeEnum.default
    context_agent_type: Union[ContextAgentTypeEnum, Callable] = ContextAgentTypeEnum.nodes_num

    selection_types: Sequence[SelectionTypesEnum] = \
        (SelectionTypesEnum.tournament,)
    crossover_types: Sequence[Union[CrossoverTypesEnum, Any]] = \
        (CrossoverTypesEnum.one_point,)
    mutation_types: Sequence[Union[MutationTypesEnum, Any]] = simple_mutation_set
    elitism_type: ElitismTypesEnum = ElitismTypesEnum.keep_n_best
    regularization_type: RegularizationTypesEnum = RegularizationTypesEnum.none
    genetic_scheme_type: GeneticSchemeTypesEnum = GeneticSchemeTypesEnum.generational

    decaying_factor: float = 1.0
    window_size: Optional[int] = None

    def __post_init__(self):
        if self.multi_objective:
            self.selection_types = (SelectionTypesEnum.spea2,)
            # TODO add possibility of using regularization in MO alg
            self.regularization_type = RegularizationTypesEnum.none



import os
from statistics import mean

from typing import Dict, List, Tuple, Any, Callable, Union, Optional

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from examples.adaptive_optimizer.utils import plot_action_values
from golem.core.log import default_log
from golem.core.optimisers.adaptive.operator_agent import OperatorAgent, ObsType
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.visualisation.opt_history.multiple_fitness_line import MultipleFitnessLines


class ExperimentAnalyzer:
    """ Class to analyze results of experiment.
    The example of usage can be found here: ~/GOLEM/examples/experiment_analyzer/
    To use this class folder hierarchy must be organized as following:

    setup (e.g. configuration of framework)
        \
        dataset
               \
               launch
                     \
                      all collected data (metrics.csv, history, saved_pipeline, etc)

    :param path_to_root: path to dir with experiment setups.
    :param folders_to_ignore: folders without experiment data to ignore.
    """

    def __init__(self, path_to_root: str, folders_to_ignore: Optional[List[str]] = None):
        if folders_to_ignore is None:
            folders_to_ignore = []
        self.path_to_root = path_to_root
        self._folders_to_ignore = folders_to_ignore
        self._log = default_log('ExperimentAnalyzer')

    def analyze_convergence(self, history_folder: Optional[str] = 'history', is_mean: bool = False,
                            path_to_save: Optional[str] = None, is_raise: bool = False,
                            metrics_names: Optional[List[str]] = None) \
            -> Dict[str, Dict[str, Dict[str, List[float]]]]:
        """ Method to analyze convergence with the use of histories.

        :param history_folder: name of the history folder in experiment result folder (e.g. 'history', 'histories').
        If history is not in separate folder than it must be specified as None.
        :param is_mean: bool flag to specify just storing all the results or calculating mean values
        :param path_to_save: path to dir where to save results
        :param is_raise: bool specifying if exception must be raised if there is no history folder
        :param metrics_names: names of metrics to consider. Useful in case of multimodal optimization.
        """
        if path_to_save:
            os.makedirs(path_to_save, exist_ok=True)

        if metrics_names is None:
            metrics_names = ['metric']

        convergence = dict()
        for setup, dataset, path_to_launch in self._get_path_to_launch():

            for metric in metrics_names:
                convergence.setdefault(metric, dict())
                convergence[metric] = \
                    self._extend_result_dict(result_dict=convergence[metric], setup=setup, dataset=dataset)

            path_to_history_folder = os.path.join(path_to_launch, history_folder) if history_folder else path_to_launch
            if not self._check_if_file_or_folder_present(path=path_to_history_folder, is_raise=is_raise):
                continue

            history_files = [file for file in os.listdir(path_to_history_folder) if file.endswith('.json')]

            # if there is no history
            if len(history_files) == 0:
                continue

            # load the first history in the folder
            history = OptHistory.load(os.path.join(path_to_history_folder, history_files[0]))

            convergence_per_metric = self._analyze_convergence(history=history,
                                                               metrics_num=len(metrics_names))
            for i, metric in enumerate(metrics_names):
                convergence[metric][setup][dataset].append(convergence_per_metric[i])

        if is_mean:
            for metric in convergence.keys():
                for setup in convergence[metric].keys():
                    for dataset in convergence[metric][setup].keys():
                        convergence[metric][setup][dataset] = mean(convergence[metric][setup][dataset])

        # save results per metric
        if path_to_save:
            for metric in metrics_names:
                df = pd.DataFrame(convergence[metric])
                path_to_save = os.path.join(path_to_save, f'convergence_results_{metric}.csv')
                df.to_csv(path_to_save)
                self._log.info(f"Convergence table was saved to {path_to_save}")
        return convergence

    @staticmethod
    def _analyze_convergence(history: OptHistory, metrics_num: int) -> List[float]:
        """ Method to get time in what individual with the best fitness was firstly obtained. """
        # find best final metric for each objective
        best_fitness_per_objective = [np.inf] * metrics_num
        for i in range(metrics_num):
            for ind in history.final_choices.data:
                if ind.fitness.values[i] < best_fitness_per_objective[i]:
                    best_fitness_per_objective[i] = ind.fitness.values[i]

        total_time_to_get_best_fitness_per_objective = []
        for j, best_fitness in enumerate(best_fitness_per_objective):
            first_gen_with_best_fitness = history.generations_count
            # since if objective is not multi-modal than historical_fitness[j] is 1d list
            historical_fitness = history.historical_fitness[j]
            if not history.objective.is_multi_objective:
                historical_fitness = [history.historical_fitness[j]]

            for i, gen_fitnesses in enumerate(historical_fitness):
                if best_fitness in gen_fitnesses:
                    first_gen_with_best_fitness = i
                    break
            total_time_to_get_best_fitness = 0
            for i, gen in enumerate(history.individuals):
                if i == first_gen_with_best_fitness:
                    break
                for ind in gen.data:
                    total_time_to_get_best_fitness += ind.metadata['computation_time_in_seconds']
            total_time_to_get_best_fitness_per_objective.append(total_time_to_get_best_fitness)
        return total_time_to_get_best_fitness_per_objective

    def analyze_metrics(self, metric_names: List[str], file_name: str, is_mean: bool = False,
                        path_to_save: Optional[str] = None, is_raise: bool = False, keep_n_best: Optional[int] = 1) \
            -> Dict[str, Dict[str, Dict[str, Union[List[float], float]]]]:
        """ Method to analyze specified metrics.
        :param metric_names: names of metrics to analyze. e.g. ['f1', 'inference_time']
        :param file_name: name of the file with metrics (e.g. 'metrics.csv').
        The file with metrics must have metric names in columns.
        :param is_mean: bool flag to specify just storing all the results or calculating mean values
        :param path_to_save: path to dir where to save results
        :param is_raise: bool specifying if exception must be raised if there is no history folder
        :param keep_n_best: number of result to consider when calculating metric.
        """
        if path_to_save:
            os.makedirs(path_to_save, exist_ok=True)

        dict_with_metrics = dict()
        for setup, dataset, path_to_launch in self._get_path_to_launch():

            df_metrics = self._get_metrics_df_from_path(path=path_to_launch, file_name=file_name, is_raise=is_raise)
            if df_metrics.empty:
                continue

            for metric in metric_names:
                dict_with_metrics.setdefault(metric, dict())
                dict_with_metrics[metric] = self._extend_result_dict(result_dict=dict_with_metrics[metric],
                                                                     setup=setup, dataset=dataset)
                if metric not in df_metrics.columns:
                    self._log.warning(f"There is no column in {file_name} with {metric}")

                # if there were multiple metrics in experiment or not
                all_metric_values = list(df_metrics[metric])
                all_metric_values.sort()
                dict_with_metrics[metric][setup][dataset].append(mean(all_metric_values[:keep_n_best]))

        if is_mean:
            for metric in metric_names:
                for setup in dict_with_metrics[metric].keys():
                    for dataset in dict_with_metrics[metric][setup].keys():
                        dict_with_metrics[metric][setup][dataset] = mean(dict_with_metrics[metric][setup][dataset])

        # save results per metric
        if path_to_save:
            for metric in dict_with_metrics.keys():
                df = pd.DataFrame(dict_with_metrics[metric])
                cur_path_to_save = os.path.join(path_to_save, f'{metric}_results.csv')
                df.to_csv(cur_path_to_save)
                self._log.info(f"Metric table was saved to {cur_path_to_save}")
        return dict_with_metrics

    def plot_convergence(self, path_to_save: str, with_confidence: bool = True,
                         history_folder: str = 'history', is_raise: bool = False,
                         metrics_names: Optional[List[str]] = None):
        """ Method to analyze convergence with the use of histories.

        :param path_to_save: path to dir where to save results
        :param with_confidence: bool param specifying to use confidence interval or not.
        :param history_folder: name of the history folder in experiment result folder (e.g. 'history', 'histories')
        :param is_raise: bool specifying if exception must be raised if there is no history folder.
        :param metrics_names: names of metrics to consider. Useful in case of multimodal optimization.
        """
        if path_to_save:
            os.makedirs(path_to_save, exist_ok=True)

        if not metrics_names:
            metrics_names = ['metric']

        histories = dict()
        for setup, dataset, path_to_launch in self._get_path_to_launch():
            histories = self._extend_result_dict(result_dict=histories, setup=setup, dataset=dataset)

            path_to_history_folder = os.path.join(path_to_launch, history_folder)
            if not self._check_if_file_or_folder_present(path=path_to_history_folder, is_raise=is_raise):
                continue

            history_files = [file for file in os.listdir(path_to_history_folder) if file.endswith('.json')]

            # if there is no history
            if len(history_files) == 0:
                continue

            # load the first history in the folder
            history = OptHistory.load(os.path.join(path_to_history_folder, history_files[0]))
            self._log.info(f"Loaded {history_files[0]} history for {setup} and {dataset} dataset")
            histories[setup][dataset].append(history.historical_fitness)

        histories_to_compare = dict()
        # plot convergence pics
        for dataset in histories[list(histories.keys())[0]]:
            for setup in histories.keys():
                histories_to_compare[setup] = histories[setup][dataset] if dataset in histories[setup].keys() else []
            multiple_fitness_plot = MultipleFitnessLines(historical_fitnesses=histories_to_compare,
                                                         metric_names=metrics_names)
            for i, metric in enumerate(metrics_names):
                file_name = f'{dataset}_convergence_with_confidence_for_{metric}' if with_confidence \
                    else f'{dataset}_convergence_without_confidence_for_{metric}'
                cur_path_to_save = os.path.join(path_to_save, file_name)
                multiple_fitness_plot.visualize(save_path=cur_path_to_save,
                                                with_confidence=with_confidence, metric_id=i)
                self._log.info(f"Convergence plot for {dataset} dataset for {metric} was saved to {cur_path_to_save}")

    def analyze_statistical_significance(self, data_to_analyze: Dict[str, Dict[str, List[float]]],
                                         stat_tests: List[Callable], path_to_save: Optional[str] = None,
                                         test_format: Optional[List[str]] = None) -> Dict[str, Dict[str, Dict[str, float]]]:
        """ Method to perform statistical analysis of data. Metric data obtained with 'analyze_metrics' and
        convergence data obtained with 'analyze_convergence' can be simply analyzed, for example.
        :param data_to_analyze: data to analyze.
        NB! data must have the specified format Dict[str, Dict[str, float]]:
        first key -- framework/setup name, second -- dataset name and then list of metric values
        :param stat_tests: list of functions of statistical tests to perform. E.g. scipy.stats.kruskal
        :param path_to_save: path to dir where to save results
        :param test_format: argument to specify what every test function must return. default: ['statistic', 'pvalue']
        """
        if path_to_save:
            os.makedirs(path_to_save, exist_ok=True)

        if not test_format:
            test_format = ['statistic', 'pvalue']

        stat_dict = dict.fromkeys(test_format, None)
        datasets = list(data_to_analyze[list(data_to_analyze.keys())[0]].keys())
        for dataset in datasets:
            values_to_compare = []
            for setup in data_to_analyze.keys():
                values_to_compare.append(data_to_analyze[setup][dataset])
            for test in stat_tests:
                try:
                    cur_test_result = test(*values_to_compare)
                except Exception as e:
                    self._log.critical(f"Statistical test ({test}) failed with exception: {e}")
                    cur_test_result = [None] * len(test_format)
                for i, arg in enumerate(test_format):
                    if not stat_dict[arg]:
                        stat_dict[arg] = dict.fromkeys([t.__name__ for t in stat_tests], None)
                    if not stat_dict[arg][test.__name__]:
                        stat_dict[arg][test.__name__] = dict.fromkeys(datasets, None)
                    stat_dict[arg][test.__name__][dataset] = cur_test_result[i]

        # save results for all tests
        if path_to_save:
            for arg in test_format:
                df = pd.DataFrame(stat_dict[arg])
                cur_path_to_save = os.path.join(path_to_save, f'stat_{arg}_results.csv')
                df.to_csv(cur_path_to_save)
                self._log.info(f"Stat test table for {arg} was saved to {cur_path_to_save}")
        return stat_dict

    def analyze_structural_complexity(self, path_to_save: str, dir_name: str, class_to_load: Any = None,
                                      load_func: Optional[Callable] = None, is_raise: bool = False,
                                      file_name: Optional[str] = None, metrics_to_display: List[str] = None):
        """ Method to save pictures of final graphs in directories to compare it visually.
        :param path_to_save: root path to pictures per setup.
        :param dir_name: name of directory in which final graph is saved.
        :param class_to_load: class of objects to load.
        :param load_func: function that load object. Can be used in case method 'load' is not defined for the object.
        :param is_raise: bool specifying if exception must be raised if there is no history folder.
        :param file_name: name of the file with metrics (e.g. 'metrics.csv')
        :param metrics_to_display: list of metrics to display in the title of the picture with result.
        """
        if path_to_save:
            os.makedirs(path_to_save, exist_ok=True)

        for setup, dataset, path_to_launch in self._get_path_to_launch():
            if not self._check_if_file_or_folder_present(path=os.path.join(path_to_launch, dir_name), is_raise=is_raise):
                continue

            path_to_json = None
            for address, dirs, files in os.walk(path_to_launch):
                for name in files:
                    if '.json' in name:
                        path_to_json = os.path.join(address, name)
                        break

            # final result was not saved in this launch
            if not path_to_json:
                continue

            result = class_to_load.load(path_to_json) if class_to_load is not None \
                else load_func(path_to_json) if load_func else None

            # load metrics for title if specified
            df_metrics = self._get_metrics_df_from_path(path=path_to_launch, file_name=file_name, is_raise=is_raise)
            title = ''
            for metric in metrics_to_display:
                if metric not in df_metrics.columns:
                    self._log.warning(f"There is no column in {file_name} with {metric}")
                else:
                    title += f'{metric}={df_metrics[metric][0]} '
            title = 'Best metrics for launch: ' + title
            cur_path_to_save = os.path.join(path_to_save, setup)

            if not os.path.exists(cur_path_to_save):
                os.makedirs(cur_path_to_save)
            saved_results = [int(cur_name.split("_")[0]) for cur_name in os.listdir(cur_path_to_save)
                             if cur_name not in self._folders_to_ignore]
            max_saved_num = max(saved_results) if saved_results else 0
            cur_path_to_save = os.path.join(cur_path_to_save, f'{max_saved_num+1}_result.png')
            result.show(cur_path_to_save, title=title)
            self._log.info(f"Resulting graph was saved to {cur_path_to_save}")

    def plot_agent_mutation_probs_and_expectations(self, path_to_save: str, agent_class: type(OperatorAgent),
                                                   mutation_names: List[str], gen_to_save_agent: int = 5,
                                                   dir_name: str = 'agent', is_raise: bool = False,
                                                   obs: Optional[ObsType] = None):
        """ Analyzes agent mutation probabilities and actions expectations through the evolution process.
        Useful for simple MAB and contextual MABs.
        :param path_to_save:
        :param agent_class: class of agent to load
        :param mutation_names: mutations that were applied by agent
        :param gen_to_save_agent: the agent is saved every generation which % gen_to_save_agent == 0.
        Originally, the agent is saved every 5th generation to not overload memory.
        :param dir_name: name of directory in which agent are saved.
        :param is_raise: bool specifying if exception must be raised if there is no history folder.
        :param obs: observation to throw to agent. E.g. ContextualMultiArmedBanditAgent needs an
        observation to predict action probabilities. If the specified agent does not need observation,
        can be left as None.
        """
        if path_to_save:
            os.makedirs(path_to_save, exist_ok=True)

        probs_dict = dict()
        for setup, dataset, path_to_launch in self._get_path_to_launch():
            if not self._check_if_file_or_folder_present(path=os.path.join(path_to_launch, dir_name),
                                                         is_raise=is_raise):
                continue

            probs_dict = self._extend_result_dict(result_dict=probs_dict, setup=setup, dataset=dataset)

            path_to_agents = os.path.join(path_to_launch, dir_name)
            cur_probabilities_list = []
            for agent_num in os.listdir(path_to_agents):
                path_cur_agent = os.path.join(path_to_agents, agent_num)
                agent = agent_class.load(path_cur_agent)
                probs = agent.get_action_values(obs)
                cur_probabilities_list.append(probs)
            probs_dict[setup][dataset].append(cur_probabilities_list)
            self._log.info(f"Agents for {setup} {dataset} were analyzed")

        # Average out probabilities per generation
        num_mutations = len(mutation_names)
        average_probabilities = {}
        for setup in probs_dict.keys():
            for dataset in probs_dict[setup].keys():
                average_probs = []
                all_probabilities = probs_dict[setup][dataset]
                max_gen = max([len(agent) for agent in all_probabilities])
                for gen in range(max_gen):
                    all_probs_per_gen = []
                    for i in range(num_mutations):
                        all_probs_per_gen.append([])
                    for agent in all_probabilities:
                        for mutation_num in range(num_mutations):
                            # extend mutations as if all launches are of the same length
                            if gen >= len(agent):
                                all_probs_per_gen[mutation_num].append(agent[-1][mutation_num])
                            else:
                                all_probs_per_gen[mutation_num].append(agent[gen][mutation_num])
                    average_probs_per_gen = []
                    for probs_list in all_probs_per_gen:
                        average_probs_per_gen.append(mean(probs_list))

                    for i in range(gen_to_save_agent):
                        average_probs.append(average_probs_per_gen)
                average_probabilities = self._extend_result_dict(result_dict=average_probabilities,
                                                                 setup=setup, dataset=dataset)
                average_probabilities[setup][dataset] = average_probs

                plot_action_values(stats=average_probabilities[setup][dataset], action_tags=mutation_names,
                                   titles=['Average action Expectation Values', 'Average action Probabilities'])
                cur_path_to_save = os.path.join(path_to_save, f'{setup}_{dataset}_probs.png')
                plt.savefig(cur_path_to_save)
                self._log.info(f"Plot of action expectations and mutation probabilities was saved: {cur_path_to_save}")

    def _get_path_to_launch(self) -> Tuple[str, str, str]:
        """ Yields setup name, dataset name and paths to dirs with experiment results.
        If experiment saving configuration/files structure somehow differs from the structure implied in this class
        this method can be used externally to get paths to launches.
        """
        for setup in os.listdir(self.path_to_root):
            if setup in self._folders_to_ignore:
                continue
            path_to_setup = os.path.join(self.path_to_root, setup)
            for dataset in os.listdir(path_to_setup):
                if dataset in self._folders_to_ignore:
                    continue
                path_to_dataset = os.path.join(path_to_setup, dataset)
                for launch in os.listdir(path_to_dataset):
                    path_to_launch = os.path.join(path_to_dataset, launch)
                    yield setup, dataset, path_to_launch

    @staticmethod
    def _extend_result_dict(result_dict: dict, setup: str, dataset: str) -> Dict[str, Dict[str, list]]:
        """ Extends result dictionary with new setup and dataset name. """
        result_dict.setdefault(setup, dict())
        result_dict[setup].setdefault(dataset, list())
        return result_dict

    def _get_metrics_df_from_path(self, path: str, file_name: str, is_raise: bool) -> pd.DataFrame:
        path_to_df = os.path.join(path, file_name)
        if not self._check_if_file_or_folder_present(path_to_df, is_raise):
            return pd.DataFrame()

        df_metrics = pd.read_csv(path_to_df)
        return df_metrics

    @staticmethod
    def _check_if_file_or_folder_present(path: str, is_raise: bool) -> bool:
        if not (os.path.isdir(path) or os.path.isfile(path)):
            if is_raise:
                raise ValueError(f"There is no folder/file with name {os.path.split(path)[-1]}")
            else:
                return False
        return True



import os
import tarfile

import matplotlib.pyplot as plt
from scipy.stats import mannwhitneyu, kruskal, ttest_ind

from experiments.experiment_analyzer import ExperimentAnalyzer
from golem.core.paths import project_root


if __name__ == '__main__':
    """ The result of analysis can be seen without running the script in
    '~/GOLEM/examples/experiment_analyzer/result_analysis.tar.gz'. """

    path_to_root = os.path.join(project_root(), 'examples', 'experiment_analyzer')

    # extract data if there is an archive
    if 'data.tar.gz' in os.listdir(path_to_root):
        tar = tarfile.open(os.path.join(path_to_root, 'data.tar.gz'), "r:gz")
        tar.extractall()
        tar.close()

    path_to_experiment_data = os.path.join(path_to_root, 'data')
    path_to_save = os.path.join(path_to_root, 'result_analysis')

    analyzer = ExperimentAnalyzer(path_to_root=path_to_experiment_data, folders_to_ignore=['result_analysis',
                                                                                           'Thumbs.db'])

    # to get convergence table with mean values
    path_to_save_convergence = os.path.join(path_to_save, 'convergence')

    convergence_mean = analyzer.analyze_convergence(history_folder='histories', is_raise=False,
                                                    path_to_save=path_to_save_convergence,
                                                    is_mean=True)

    # to get convergence boxplots
    convergence = analyzer.analyze_convergence(history_folder='histories', is_raise=False)
    path_to_save_convergence_boxplots = os.path.join(path_to_save_convergence, 'convergence_boxplots')

    metrics = list(convergence.keys())
    setups = list(convergence[metrics[0]].keys())
    datasets = list(convergence[metrics[0]][setups[0]].keys())
    for dataset in datasets:
        for metric_name in convergence.keys():
            to_compare = dict()
            for setup in convergence[metric_name].keys():
                to_compare[setup] = [i for i in convergence[metric_name][setup][dataset]]
            plt.boxplot(list(to_compare.values()), labels=list(to_compare.keys()))
            plt.title(f'Convergence on {dataset}')
            os.makedirs(path_to_save_convergence_boxplots, exist_ok=True)
            plt.savefig(os.path.join(path_to_save_convergence_boxplots, f'convergence_{dataset}.png'))
            plt.close()

    # to get metrics table with mean values
    path_to_save_metrics = os.path.join(path_to_save, 'metrics')
    metric_names = ['roc_auc', 'f1']
    metrics_dict_mean = analyzer.analyze_metrics(metric_names=metric_names, file_name='evaluation_results.csv',
                                                 is_raise=False, path_to_save=path_to_save_metrics,
                                                 is_mean=True)

    # to get metrics boxplots
    metrics_dict = analyzer.analyze_metrics(metric_names=metric_names, file_name='evaluation_results.csv',
                                            is_raise=False)
    path_to_save_metrics_boxplots = os.path.join(path_to_save_metrics, 'metrics_boxplot')

    for metric in metric_names:
        for dataset in metrics_dict[metric][list(metrics_dict[metric].keys())[0]].keys():
            to_compare = dict()
            for setup in metrics_dict[metric].keys():
                to_compare[setup] = [-1 * i for i in metrics_dict[metric][setup][dataset]]
            plt.boxplot(list(to_compare.values()), labels=list(to_compare.keys()))
            plt.title(f'{metric} on {dataset}')
            cur_path_to_save = os.path.join(path_to_save_metrics_boxplots, metric)
            os.makedirs(cur_path_to_save, exist_ok=True)
            plt.savefig(os.path.join(cur_path_to_save, f'{metric}_{dataset}.png'))
            plt.close()

    # to get stat test results table
    path_to_save_stat = os.path.join(path_to_save, 'statistic')
    stat_dict = analyzer.analyze_statistical_significance(data_to_analyze=metrics_dict['roc_auc'],
                                                          stat_tests=[mannwhitneyu, kruskal, ttest_ind],
                                                          path_to_save=path_to_save_stat)



from typing import Sequence

import networkx as nx
import numpy as np

from golem.metrics.graph_features import degree_stats
from libs.netcomp import _eigs, normalized_laplacian_eig
from libs.netcomp import laplacian_matrix


def nxgraph_stats(graph: nx.Graph):
    degrees = nx.degree_histogram(graph)
    degrees_norm = np.divide(degrees, np.linalg.norm(degrees)).round(2)
    stats = dict(
        num_nodes=graph.number_of_nodes(),
        num_edges=graph.number_of_edges(),
        avg_clustering=np.round(nx.average_clustering(graph), 3),
        degrees_hist=degrees,
        degrees_hist_norm=degrees_norm,
    )
    return stats


def degree_distance_kernel(target_graph: nx.DiGraph, graph: nx.DiGraph) -> float:
    return degree_stats([graph], [target_graph])


def degree_distance(target_graph: nx.DiGraph,
                    graph: nx.DiGraph,
                    normalized: bool = False) -> float:
    """This is a heuristic metric for graphs where central
    nodes are more important than peripheral ones. The "heavier"
    the nodes (i.e. the higher their degree), the more significant
    the difference in number of such nodes between two graphs."""
    # Compute histogram of node degrees
    degrees_t = np.array(nx.degree_histogram(target_graph), dtype=float)
    degrees_g = np.array(nx.degree_histogram(graph), dtype=float)
    return degree_dist_weighted_compute(degrees_t, degrees_g, normalized)


def degree_dist_weighted_compute(degrees_t: Sequence[float],
                                 degrees_g: Sequence[float],
                                 normalized: bool = False) -> float:
    degrees_t = np.asarray(degrees_t)
    degrees_g = np.asarray(degrees_g)

    # Extend arrays to the same length with zeros
    common_len = max(len(degrees_t), len(degrees_g))
    degrees_t.resize(common_len, refcheck=False)
    degrees_g.resize(common_len, refcheck=False)

    # Compute weights as normalized degrees
    weights = np.arange(1, common_len + 1).astype(float)
    weights /= np.sum(weights)

    # Normalize
    if normalized:
        degrees_t /= np.sum(degrees_t)
        degrees_g /= np.sum(degrees_g)

    # Compute distance between node degrees weighted by degree
    dist = np.linalg.norm(weights * (degrees_t - degrees_g))
    return dist


def size_diff(target_graph: nx.DiGraph, graph: nx.DiGraph) -> float:
    nodes_diff = abs(target_graph.number_of_nodes() - graph.number_of_nodes())
    edges_diff = abs(target_graph.number_of_edges() - graph.number_of_edges())
    return nodes_diff + np.sqrt(edges_diff)


def spectral_dist(target_graph: nx.DiGraph, graph: nx.DiGraph,
                  k: int = 20, kind: str = 'laplacian',
                  size_diff_penalty: float = 0.2,
                  match_size: bool = False,
                  ) -> float:
    target_adj = nx.adjacency_matrix(target_graph)
    adj = nx.adjacency_matrix(graph)

    # compute spectral distance
    value = lambda_dist(target_adj, adj, kind=kind, k=k, match_size=match_size)

    if size_diff_penalty > 1e-5:
        value += size_diff_penalty * size_diff(target_graph, graph)
    return value


def spectral_dists_all(target_graph: nx.DiGraph, graph: nx.DiGraph,
                       k: int = 20, match_size: bool = True) -> dict:
    target_adj = nx.adjacency_matrix(target_graph)
    adj = nx.adjacency_matrix(graph)

    print(f'computing metrics for {k} spectral values between {target_adj.shape} & {adj.shape}')

    vals = {}
    for kind in ('adjacency', 'laplacian_norm', 'laplacian'):
        value = lambda_dist(target_adj, adj, kind=kind, k=k, match_size=match_size)
        vals[kind] = np.round(value, 3)
    vals['nodes_diff'] = size_diff(target_graph, graph)
    return vals


def min_max(a, b):
    return (a, b) if a <= b else (b, a)


def lambda_dist(A1, A2, k=None, p=2, kind='laplacian', match_size=True):
    """The lambda distance between graphs, which is defined as

        d(G1,G2) = norm(L_1 - L_2)

    where L_1 is a vector of the top k eigenvalues of the appropriate matrix
    associated with G1, and L2 is defined similarly.

    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared

    k : Integer
        The number of eigenvalues to be compared

    p : non-zero Float
        The p-norm is used to compare the resulting vector of eigenvalues.

    kind : String , in {'laplacian','laplacian_norm','adjacency'}
        The matrix for which eigenvalues will be calculated.

    Returns
    -------
    dist : float
        The distance between the two graphs

    Notes
    -----
    The norm can be any p-norm; by default we use p=2. If p<0 is used, the
    result is not a mathematical norm, but may still be interesting and/or
    useful.

    If k is provided, then we use the k SMALLEST eigenvalues for the Laplacian
    distances, and we use the k LARGEST eigenvalues for the adjacency
    distance. This is because the corresponding order flips, as L = D-A.

    References
    ----------

    See Also
    --------
    netcomp.linalg._eigs
    normalized_laplacian_eigs

    """
    # check sizes & determine number of eigenvalues (k)
    nmin, nmax = min_max(A1.shape[0], A2.shape[0])
    if match_size:
        shape = (nmax, nmax)
        A1.resize(shape)
        A2.resize(shape)
    else:
        k = min(k, nmin)

    if kind == 'laplacian':
        # form matrices
        L1, L2 = [laplacian_matrix(A) for A in [A1, A2]]
        # get eigenvalues, ignore eigenvectors
        evals1, evals2 = [_eigs(L)[0] for L in [L1, L2]]
    elif kind == 'laplacian_norm':
        # use our function to graph evals of normalized laplacian
        evals1, evals2 = [normalized_laplacian_eig(A)[0] for A in [A1, A2]]
    elif kind == 'adjacency':
        evals1, evals2 = [_eigs(A)[0] for A in [A1, A2]]
        # reverse, so that we are sorted from large to small, since we care
        # about the k LARGEST eigenvalues for the adjacency distance
        evals1, evals2 = [evals[::-1] for evals in [evals1, evals2]]
    else:
        raise AttributeError(f"Invalid type {kind}, choose from 'laplacian', "
                             f"'laplacian_norm', and 'adjacency'.")
    dist = np.linalg.norm(evals1[:k] - evals2[:k], ord=p)
    return dist



import copy
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

"""Code from https://github.com/mpcrlab/MolecularTransformerEmbeddings"""

PRINTABLE_ASCII_CHARS = 95

_extra_chars = ["seq_start", "seq_end", "pad"]
EXTRA_CHARS = {key: chr(PRINTABLE_ASCII_CHARS + i) for i, key in enumerate(_extra_chars)}
ALPHABET_SIZE = PRINTABLE_ASCII_CHARS + len(EXTRA_CHARS)


class Embedding(nn.Module):
    def __init__(self, alphabet_size, d_model):
        super().__init__()
        self.alphabet_size = alphabet_size
        self.d_model = d_model
        self.embed = nn.Embedding(alphabet_size, d_model)

    def forward(self, x):
        return self.embed(x)


class PositionalEncoder(nn.Module):
    def __init__(self, d_model, max_seq_len=6000, dropout=0.1):
        super().__init__()
        self.d_model = d_model
        self.dropout = nn.Dropout(p=dropout)
        # create constant 'pe' matrix with values dependant on pos and i
        pe = torch.zeros(max_seq_len, d_model)
        for pos in range(max_seq_len):
            for i in range(0, d_model, 2):
                pe[pos, i] = math.sin(pos / (10000 ** ((2 * i) / d_model)))
                pe[pos, i + 1] = math.cos(pos / (10000 ** ((2 * (i + 1)) / d_model)))
        pe = pe.unsqueeze(0)
        self.register_buffer('pe', pe)

    def forward(self, x):
        # make embeddings relatively larger
        x = x * math.sqrt(self.d_model)
        # add constant to embedding
        seq_len = x.size(1)
        pe = Variable(self.pe[:, :seq_len], requires_grad=False)
        if x.is_cuda:
            pe.cuda()
        x = x + pe
        x = self.dropout(x)
        return x


class Norm(nn.Module):
    def __init__(self, d_model, eps=1e-6):
        super().__init__()
        self.size = d_model
        # create two learnable parameters to calibrate normalisation
        self.alpha = nn.Parameter(torch.ones(self.size))
        self.bias = nn.Parameter(torch.zeros(self.size))
        self.eps = eps

    def forward(self, x):
        norm = self.alpha * (x - x.mean(dim=-1, keepdim=True)) / (x.std(dim=-1, keepdim=True) + self.eps) + self.bias
        return norm


def attention(q, k, v, d_k, mask=None, dropout=None):
    scores = torch.matmul(q, k.transpose(-2, -1)) / math.sqrt(d_k)
    if mask is not None:
        mask = mask.unsqueeze(1)
        scores = scores.masked_fill(mask == 0, -1e9)
    scores = F.softmax(scores, dim=-1)
    if dropout is not None:
        scores = dropout(scores)
    output = torch.matmul(scores, v)
    return output


class MultiHeadAttention(nn.Module):
    def __init__(self, heads, d_model, dropout=0.1):
        super().__init__()
        self.d_model = d_model
        self.d_k = d_model // heads
        self.h = heads
        self.q_linear = nn.Linear(d_model, d_model)
        self.v_linear = nn.Linear(d_model, d_model)
        self.k_linear = nn.Linear(d_model, d_model)
        self.dropout = nn.Dropout(dropout)
        self.out = nn.Linear(d_model, d_model)

    def forward(self, q, k, v, mask=None):
        bs = q.size(0)
        # perform linear operation and split into N heads
        k = self.k_linear(k).view(bs, -1, self.h, self.d_k)
        q = self.q_linear(q).view(bs, -1, self.h, self.d_k)
        v = self.v_linear(v).view(bs, -1, self.h, self.d_k)
        # transpose to get dimensions bs * N * sl * d_model
        k = k.transpose(1, 2)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)
        # calculate attention using function we will define next
        scores = attention(q, k, v, self.d_k, mask, self.dropout)
        # concatenate heads and put through final linear layer
        concat = scores.transpose(1, 2).contiguous().view(bs, -1, self.d_model)
        output = self.out(concat)
        return output


class FeedForward(nn.Module):
    def __init__(self, d_model, d_ff=2048, dropout=0.1):
        super().__init__()
        # We set d_ff as a default to 2048
        self.linear_1 = nn.Linear(d_model, d_ff)
        self.dropout = nn.Dropout(dropout)
        self.linear_2 = nn.Linear(d_ff, d_model)

    def forward(self, x):
        x = self.dropout(F.relu(self.linear_1(x)))
        x = self.linear_2(x)
        return x


class EncoderLayer(nn.Module):
    def __init__(self, d_model, heads, dropout=0.1):
        super().__init__()
        self.norm_1 = Norm(d_model)
        self.norm_2 = Norm(d_model)
        self.attn = MultiHeadAttention(heads, d_model, dropout=dropout)
        self.ff = FeedForward(d_model, dropout=dropout)
        self.dropout_1 = nn.Dropout(dropout)
        self.dropout_2 = nn.Dropout(dropout)

    def forward(self, x, mask):
        x2 = self.norm_1(x)
        x = x + self.dropout_1(self.attn(x2, x2, x2, mask))
        x2 = self.norm_2(x)
        x = x + self.dropout_2(self.ff(x2))
        return x


# build a decoder layer with two multi-head attention layers and
# one feed-forward layer
class DecoderLayer(nn.Module):
    def __init__(self, d_model, heads, dropout=0.1):
        super().__init__()
        self.norm_1 = Norm(d_model)
        self.norm_2 = Norm(d_model)
        self.norm_3 = Norm(d_model)

        self.dropout_1 = nn.Dropout(dropout)
        self.dropout_2 = nn.Dropout(dropout)
        self.dropout_3 = nn.Dropout(dropout)

        self.attn_1 = MultiHeadAttention(heads, d_model, dropout=dropout)
        self.attn_2 = MultiHeadAttention(heads, d_model, dropout=dropout)
        self.ff = FeedForward(d_model, dropout=dropout)

    def forward(self, x, e_outputs, src_mask, trg_mask):
        x2 = self.norm_1(x)
        x = x + self.dropout_1(self.attn_1(x2, x2, x2, trg_mask))
        x2 = self.norm_2(x)
        x = x + self.dropout_2(self.attn_2(x2, e_outputs, e_outputs, src_mask))
        x2 = self.norm_3(x)
        x = x + self.dropout_3(self.ff(x2))
        return x


def get_clones(module, n):
    return nn.ModuleList([copy.deepcopy(module) for _ in range(n)])


class Encoder(nn.Module):
    def __init__(self, alphabet_size, d_model, n, heads, dropout):
        super().__init__()
        self.n = n
        self.embed = Embedding(alphabet_size, d_model)
        self.pe = PositionalEncoder(d_model, dropout=dropout)
        self.layers = get_clones(EncoderLayer(d_model, heads, dropout), n)
        self.norm = Norm(d_model)

    def forward(self, src, mask):
        x = self.embed(src)
        x = self.pe(x)
        for i in range(self.n):
            x = self.layers[i](x, mask)
        return self.norm(x)


class Decoder(nn.Module):
    def __init__(self, alphabet_size, d_model, n, heads, dropout):
        super().__init__()
        self.n = n
        self.embed = Embedding(alphabet_size, d_model)
        self.pe = PositionalEncoder(d_model, dropout=dropout)
        self.layers = get_clones(DecoderLayer(d_model, heads, dropout), n)
        self.norm = Norm(d_model)

    def forward(self, trg, e_outputs, src_mask, trg_mask):
        x = self.embed(trg)
        x = self.pe(x)
        for i in range(self.n):
            x = self.layers[i](x, e_outputs, src_mask, trg_mask)
        return self.norm(x)


class Transformer(nn.Module):
    def __init__(self, alphabet_size, d_model, n, heads=8, dropout=0.1):
        super().__init__()
        self.encoder = Encoder(alphabet_size, d_model, n, heads, dropout)
        self.decoder = Decoder(alphabet_size, d_model, n, heads, dropout)
        self.out = nn.Linear(d_model, alphabet_size)

    def forward(self, src, trg, src_mask, trg_mask):
        e_outputs = self.encoder(src, src_mask)
        d_output = self.decoder(trg, e_outputs, src_mask, trg_mask)
        output = self.out(d_output)
        return output


def nopeak_mask(size, device):
    np_mask = torch.triu(torch.ones((size, size), dtype=torch.uint8), diagonal=1).unsqueeze(0)
    np_mask = np_mask == 0
    np_mask = np_mask.to(device)
    return np_mask


def create_masks(src, trg=None, pad_idx=ord(EXTRA_CHARS['pad']), device=None):
    src_mask = (src != pad_idx).unsqueeze(-2)

    if trg is not None:
        trg_mask = (trg != pad_idx).unsqueeze(-2)
        size = trg.size(1)  # get seq_len for matrix
        np_mask = nopeak_mask(size, device)
        np_mask.to(device)
        trg_mask = trg_mask & np_mask
        return src_mask, trg_mask
    return src_mask



import os
import pickle
import sys
from typing import Dict

from rdkit import RDConfig, Chem
from rdkit.Chem import Descriptors, AllChem
from rdkit.Chem.QED import qed
from rdkit.Chem.rdchem import RWMol

from examples.molecule_search.constants import ZINC_LOGP_MEAN, ZINC_LOGP_STD, ZINC_SA_MEAN, ZINC_SA_STD, \
    ZINC_CYCLE_MEAN, ZINC_CYCLE_STD, MIN_LONG_CYCLE_SIZE
from examples.molecule_search.mol_graph import MolGraph
from examples.molecule_search.utils import largest_ring_size, download_from_github
from golem.core.paths import project_root

sys.path.append(os.path.join(RDConfig.RDContribDir, 'SA_Score'))
import sascorer


def qed_score(mol_graph: MolGraph):
    """Quantitative estimation of drug-likeness. QED measure reflects the underlying distribution of molecular
    properties including molecular weight, logP, topological polar surface area, number of hydrogen bond donors and
    acceptors, the number of aromatic rings and rotatable bonds, and the presence of unwanted chemical
    functionalities.

    It is a numerical value ranging from 0 to 1, with a higher score indicating a greater likelihood
    of the molecule having favorable drug-like properties.
    """

    molecule = mol_graph.get_rw_molecule(aromatic=True)
    score = qed(molecule)
    return -score


def sa_score(mol_graph: MolGraph) -> float:
    """Synthetic Accessibility score is a metric used to evaluate the ease of synthesizing a molecule.
    The SA score takes into account a variety of factors such as the number of synthetic steps,
    the availability of starting materials, and the feasibility of the reaction conditions required for each step.

    It is ranged between 1 and 10, 1 is the best possible score.
    """

    molecule = mol_graph.get_rw_molecule(aromatic=True)
    score = sascorer.calculateScore(molecule)
    return score


def penalised_logp(mol_graph: MolGraph) -> float:
    """ LogP penalized by SA and length of long cycles, as described in (Kusner et al. 2017).
    LogP (Octanol-Water Partition Coefficient) is a measure of the lipophilicity,
    or ability to dissolve in lipids, of a molecule. It is commonly used to predict the bioavailability
    and pharmacokinetic properties of drugs.

    This version is penalized by SA score and the length of the largest cycle.
    """

    molecule = mol_graph.get_rw_molecule(aromatic=True)
    log_p = Descriptors.MolLogP(molecule)
    sa_score = sascorer.calculateScore(molecule)
    largest_cycle_size = largest_ring_size(molecule)
    cycle_score = max(largest_cycle_size - MIN_LONG_CYCLE_SIZE, 0)
    score = log_p - sa_score - cycle_score
    return -score


def normalized_logp(mol_graph: MolGraph) -> float:
    """Normalized LogP based on the statistics of 250k_rndm_zinc_drugs_clean.smi dataset.

    Original code:
    https://github.com/bowenliu16/rl_graph_generation/blob/master/gym-molecule/gym_molecule/envs/molecule.py
    """
    molecule = mol_graph.get_rw_molecule(aromatic=True)

    logp = Descriptors.MolLogP(molecule)
    sa = -sascorer.calculateScore(molecule)

    largest_ring = largest_ring_size(molecule)
    cycle_score = max(largest_ring - 6, 0)

    norm_logp = (logp - ZINC_LOGP_MEAN) / ZINC_LOGP_STD
    normalized_sa = (sa - ZINC_SA_MEAN) / ZINC_SA_STD
    normalized_cycle = (cycle_score - ZINC_CYCLE_MEAN) / ZINC_CYCLE_STD

    score = norm_logp + normalized_sa + normalized_cycle

    return score


def normalized_sa_score(mol_graph: MolGraph) -> float:
    """SA score normalized to be ranged from 0 to 1, where 1 is preferable."""

    score = sa_score(mol_graph)
    normalized_score = 1 - (score - 1) / 9
    return -normalized_score


class CLScorer:
    """
    ChEMBL-likeness score (CLscore) is defined by considering which substructures in a molecule
    also occur in molecules from the public database ChEMBL, using a subset of molecules with
    reported high confidence datapoint of activity on single protein targets.

    Original code: https://github.com/reymond-group/GDBChEMBL

    Args:
        weighted: If True calculate score by summing up log10 of frequency of occurrence inside the reference database.
            When set to False, score will add 1 for every shingle inside the reference database,
            irrespective of how often it occurs. It is recommended to use the actual CLscore with weighted shingles.
        radius: Maximum radius of circular substructures around the rooting atom.
            Note that when using the ChEMBL shingle library, maximum radius will be 3 (default).
            For larger radii, the reference database would have to be read out accordingly.
        rooted: Use rooted shingles. This means reference shingles are canonical but always starting at the central
            atom of a circular substructure. False means shingles in the database are canonicalized but not rooted.
            It is recommended to use rooted shingles.
    """
    ROOTED_FILE_PATH = "examples/molecule_search/data/shingles/" \
                       "chembl_24_1_shingle_scores_log10_rooted_nchir_min_freq_100.pkl"
    NOT_ROOTED_FILE_PATH = "examples/molecule_search/data/shingles/chembl_24_1_shingle_scores_log10_nrooted_nchir.pkl"

    ROOTED_GITHUB_URL = "https://github.com/aimclub/GOLEM/raw/molecule-search/examples/molecule_search/data/shingles/" \
                        "chembl_24_1_shingle_scores_log10_rooted_nchir_min_freq_100.pkl"
    NOT_ROOTED_GITHUB_URL = "https://github.com/aimclub/GOLEM/raw/molecule-search/examples/molecule_search/data/" \
                            "shingles/chembl_24_1_shingle_scores_log10_nrooted_nchir.pkl"

    def __init__(self, weighted: bool = True, radius: int = 3, rooted: bool = True):
        self.rooted = rooted
        self.weighted = weighted
        self.radius = radius

        file_path = CLScorer.ROOTED_FILE_PATH if self.rooted else CLScorer.NOT_ROOTED_FILE_PATH
        self.file_path = os.path.join(project_root(), file_path)
        self.github_url = CLScorer.ROOTED_GITHUB_URL if self.rooted else CLScorer.NOT_ROOTED_GITHUB_URL
        self.db_shingles = self.load_shingles()

    def __call__(self, mol_graph: MolGraph) -> float:
        """
        Args:
            mol_graph: MolGraph to evaluate
        """
        molecule = mol_graph.get_rw_molecule(aromatic=True)
        qry_shingles = self.extract_shingles(molecule)

        # calculate shingle count averaged score
        avg_score = 0
        if qry_shingles:
            sum_scores = 0
            # using log10 of shingle frequency
            if self.weighted:
                for shingle in qry_shingles:
                    sum_scores += self.db_shingles.get(shingle, 0)
            # working binary (i.e. if present -> count++ )
            else:
                for shingle in qry_shingles:
                    if shingle in self.db_shingles:
                        sum_scores += 1
            avg_score = sum_scores / len(qry_shingles)

        return -avg_score

    def load_shingles(self) -> Dict:
        download_from_github(self.file_path, self.github_url)

        with open(self.file_path, "rb") as pyc:
            db_shingles = pickle.load(pyc)
        return db_shingles

    def extract_shingles(self, molecule: RWMol) -> set:
        qry_shingles = set()

        radius_constr = self.radius + 1

        for atm_idx in range(molecule.GetNumAtoms()):
            for N in range(1, radius_constr):
                bonds = AllChem.FindAtomEnvironmentOfRadiusN(molecule, N, atm_idx)

                if not bonds:
                    break

                atoms = set()
                for bond_id in bonds:
                    bond = molecule.GetBondWithIdx(bond_id)
                    atoms.add(bond.GetBeginAtomIdx())
                    atoms.add(bond.GetEndAtomIdx())

                if self.rooted:
                    new_shingle = Chem.MolFragmentToSmiles(molecule,
                                                           atomsToUse=list(atoms),
                                                           bondsToUse=bonds,
                                                           isomericSmiles=False,
                                                           rootedAtAtom=atm_idx)
                else:
                    new_shingle = Chem.MolFragmentToSmiles(molecule,
                                                           atomsToUse=list(atoms),
                                                           bondsToUse=bonds,
                                                           isomericSmiles=False,
                                                           rootedAtAtom=-1)

                qry_shingles.add(new_shingle)

        return qry_shingles



import os.path
from itertools import chain
from typing import Tuple, Optional, Sequence

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib import cm

from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.utilities.data_structures import ensure_wrapped_in_sequence
from golem.visualisation.graph_viz import GraphVisualizer


def fitness_to_stats(history: OptHistory,
                     target_metric_index: int = 0) -> Tuple[np.ndarray, np.ndarray]:
    all_metrics = history.historical_fitness
    if history.objective.is_multi_objective:
        quality = [[metrics[target_metric_index] for metrics in pop]
                   for pop in all_metrics]
    else:
        quality = all_metrics

    quality = np.array(quality)
    if quality.shape[1] == 1:
        # reshape array to get nontrivial samples for computing mean/std
        sample_size = 20
        shape = (len(quality) // sample_size, sample_size)
        num_elements = shape[0] * shape[1]
        quality = quality.squeeze()[:num_elements].reshape(shape)

    mean = np.mean(quality, axis=1)
    std = np.std(quality, axis=1)
    return mean, std


def plot_fitness_comparison(histories: Sequence[OptHistory],
                            target_metric_index: int = 0,
                            titles: Optional[Sequence[str]] = None,
                            total_minutes: Optional[float] = None,
                            ):
    stats = [fitness_to_stats(h, target_metric_index) for h in histories]

    if not titles:
        titles = [f'history_{i}' for i in range(len(histories))]
    common_len = min(len(mean) for mean, std in stats)
    xs = np.arange(0, common_len) if not total_minutes else np.linspace(0, total_minutes, common_len)

    fig, ax = plt.subplots()
    ax.set_title('Historical fitness comparison')
    ax.set_xlabel('minutes')
    ax.set_ylabel('fitness')
    ax.legend()
    fig.tight_layout()

    def resample(arr):
        old_indices = np.arange(0, len(arr))
        new_indices = np.linspace(0, len(arr), common_len)
        return np.interp(new_indices, old_indices, arr)

    for title, (mean, std) in zip(titles, stats):
        mean = resample(mean)
        std = resample(std)

        ax.plot(xs, mean, '-', label=title)
        ax.fill_between(xs, mean - std, mean + std, alpha=0.25)

    return fig, ax


def plot_nx_graph(g: nx.DiGraph, ax: plt.Axes = None):
    adapter = BaseNetworkxAdapter()
    GraphVisualizer(adapter.adapt(g)).draw_nx_dag(ax,
                                                  node_size_scale=0.2,
                                                  font_size_scale=0.25,
                                                  edge_curvature_scale=0.5)


def draw_graphs_subplots(*graphs: nx.Graph,
                         titles: Optional[Sequence[str]] = None,
                         color_degrees: bool = False,
                         with_labels: bool = False,
                         draw_fn=nx.draw,
                         size: int = 10,
                         show: bool = True,
                         path_to_save: Optional[str] = None):
    """Draw 1 or several NetworkX graphs
    with coloring of roots, sources and node degrees.

    Args:
        graphs: 1 or many instances of nx.Graph
        titles: optional sequence of titles for graph subplots
        color_degrees: flag specifying if graph nodes must be colored relative to their degree
        with_labels: flag specifying if graph nodes must be drawn with labels
        draw_fn: nx.draw_* primitive for drawing graph subplots
        size: figure size, passed to plt.subplots(figsize=...)
        show: if plt.show() is necessary
        path_to_save: path to dir where to save results
    """

    graphs = ensure_wrapped_in_sequence(graphs)
    titles = [f'Graph #{i + 1}' for i in range(len(graphs))] if not titles else titles
    # Setup subplots
    ncols = int(np.ceil(np.sqrt(len(graphs))))
    nrows = len(graphs) // ncols
    aspect = nrows / ncols
    figsize = (size, int(size * aspect))
    fig, axs = plt.subplots(nrows, ncols, figsize=figsize)
    axs = np.atleast_2d(axs)
    # Draw graphs
    for title, ax, graph in zip(titles, chain(*axs), graphs):
        colors, labeldict, legend_handles = _get_node_colors_and_labels(graph, color_degrees)
        draw_fn(graph, ax=ax, arrows=True,
                node_color=colors, with_labels=with_labels, labels=labeldict)
        ax.set_title(title)
    fig.legend(handles=legend_handles)
    if show:
        plt.show()
    if path_to_save:
        plt.savefig(os.path.join(path_to_save, 'graphs_subplots.png'))


def _get_node_colors_and_labels(graph: nx.Graph,
                                color_degrees: bool = True,
                                cmap_name='viridis'):
    degrees = dict(graph.degree())
    if isinstance(graph, nx.DiGraph):
        roots = {n for n, d in graph.out_degree() if d == 0}
        sources = {n for n, d in graph.in_degree() if d == 0}
    else:
        roots = {max(*graph.nodes(), key=lambda n: graph.degree[n])}
        sources = {min(*graph.nodes(), key=lambda n: graph.degree[n])}

    max_degree = max(2, max(degrees.values()))
    root_cm = max_degree
    node_cm = max_degree // 2
    src_cm = 0
    colormap = cm.get_cmap(cmap_name, max_degree + 1)
    colors = []
    labels = {}

    for node, data in graph.nodes(data=True):
        # Determine color of the node:
        # if node is root -- use special 'min' color
        # if node is source -- use special 'max' color
        # else use node color according to colormap and node degree
        if node in roots:
            color = colormap(root_cm)
        elif node in sources:
            color = colormap(src_cm)
        else:
            clr_id = graph.degree(node) if color_degrees else node_cm
            color = colormap(clr_id)
        colors.append(color)

        # Get node label
        label = data.get('name') or str(node)
        labels[node] = label

    # Construct legend handles
    root_legend = mpatches.Patch(color=colormap(root_cm), label='Root')
    default_legend = mpatches.Patch(color=colormap(node_cm),
                                    label='Node Degree' if color_degrees else 'Intermediate')
    source_legend = mpatches.Patch(color=colormap(src_cm), label='Source')
    handles = [root_legend, default_legend, source_legend]

    return colors, labels, handles



from copy import deepcopy
from typing import Union, Tuple, Optional, List, Dict

from golem.core.adapter.adapter import DomainStructureType
from golem.core.dag.graph_node import GraphNode


class GraphBuilder:
    """ Builder for incremental construction of directed acyclic graphs.
    Semantics:
    - Forward-only & addition-only (can't prepend or delete nodes).
    - Doesn't throw, doesn't fail: methods always have a way to interpret input given current graph state.
    - Is not responsible for the validity of resulting graph (e.g. correct order, valid operations).
    - Builds always new graphs (on copies of nodes), preserves its state between builds. State doesn't leak outside.
    """

    OperationType = Union[str, Tuple[str, dict]]

    def __init__(self, graph_adapter: Optional[DomainStructureType] = None, *initial_nodes: Optional[GraphNode]):
        """
        Create builder with prebuilt nodes as origins of the branches.
            
        Args:
            graph_adapter: adapter to adapt built graph to particular graph type.
            initial_nodes: variadic positional argument with initial graph nodes
        """
        self.graph_adapter = graph_adapter
        self.heads: List[GraphNode] = list(filter(None, initial_nodes))

    @property
    def _iend(self) -> int:
        return len(self.heads)

    def reset(self):
        """ Reset builder state. """
        self.heads = []

    def to_nodes(self) -> List[GraphNode]:
        """
        Return list of final nodes and reset internal state.
        :return: list of final nodes, possibly empty.
        """
        return deepcopy(self.heads)

    def add_node(self, operation_type: Optional[str], branch_idx: int = 0, params: Optional[Dict] = None):
        """ Add single node to graph branch of specified index.
        If there are no heads => adds single Node.
        If there is single head => adds single Node using head as input.
        If there are several heads => adds single Node using as input the head indexed by branch_idx.
        If input is None => do nothing.
        If branch_idx is out of bounds => appends new Node.

        :param operation_type: new operation, possibly None
        :param branch_idx: index of the head to use as input for the new node
        :param params: parameters dictionary for the specific operation
        :return: self
        """
        raise NotImplementedError()

    def add_sequence(self, *operation_type: OperationType, branch_idx: int = 0):
        """ Same as .node() but for many operations at once.

        :param operation_type: operations for new nodes, either as an operation name
            or as a tuple of operation name and operation parameters.
        :param branch_idx: index of the branch for branching its tip
        """
        raise NotImplementedError()

    def grow_branches(self, *operation_type: Optional[OperationType]):
        """ Add single node to each branch.

        Argument position means index of the branch to grow.
        None operation means don't grow that branch.
        If there are no nodes => creates new branches.
        If number of input nodes is bigger than number of branches => extra operations create new branches.

        :param operation_type: operations for new nodes, either as an operation name
            or as a tuple of operation name and operation parameters.
        :return: self
        """
        raise NotImplementedError()

    def add_branch(self, *operation_type: Optional[OperationType], branch_idx: int = 0):
        """ Create branches at the tip of branch with branch_idx.

        None operations are filtered out.
        Number of new branches equals to number of provided operations.
        If there are no heads => will add several nodes.
        If there is single head => add several nodes using head as the previous.
        If there are several heads => branch head indexed by branch_idx.
        If branch_idx is out of bounds => adds nodes as new heads at the end.
        If no not-None operations are provided, nothing is changed.

        :param operation_type: operations for new nodes, either as an operation name
            or as a tuple of operation name and operation parameters.
        :param branch_idx: index of the branch for branching its tip
        :return: self
        """
        raise NotImplementedError()

    def add_skip_connection_edge(self, branch_idx_first: int, branch_idx_second: int,
                                 node_idx_in_branch_first: int, node_idx_in_branch_second: int):
        """ Joins two nodes which are not placed sequential in one branch.
        Can be used only in the very end of graph building but before 'join_branches'.
        Edge is directed from the first node to the second.

        :param branch_idx_first: index of the first branch which to take the first node
        :param branch_idx_second: index of the second branch which to take the second node
        :param node_idx_in_branch_first: index of the node in its branch
        :param node_idx_in_branch_second: index of the node in its branch
        """
        raise NotImplementedError()

    def join_branches(self, operation_type: Optional[str], params: Optional[Dict] = None):
        """ Joins all current branches with provided operation as ensemble node.

        If there are no branches => does nothing.
        If there is single branch => adds single node using it as input.
        If there are several branches => adds single node using all heads as inputs.

        :param operation_type: operation to use for joined node
        :param params: parameters dictionary for the specific operation
        :return: self
        """
        raise NotImplementedError()

    @staticmethod
    def _unpack_operation(operation: Optional[OperationType]) -> Tuple[Optional[str], Optional[Dict]]:
        if isinstance(operation, str) or operation is None:
            return operation, None
        else:
            return operation

    @staticmethod
    def _pack_params(name: str, params: Optional[dict]) -> Optional[dict]:
        return {'name': name, 'params': params} if params else {'name': name}



from dataclasses import dataclass, field
from typing import List, Dict, Generic, Tuple, Any, Optional

import numpy as np
from iOpt.method.listener import ConsoleFullOutputListener
from iOpt.problem import Problem
from iOpt.solver import Solver
from iOpt.solver_parametrs import SolverParameters
from iOpt.trial import Point, FunctionValue

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveEvaluate
from golem.core.tuning.search_space import SearchSpace, get_node_operation_parameter_label
from golem.core.tuning.tuner_interface import BaseTuner, DomainGraphForTune


@dataclass
class IOptProblemParameters:
    float_parameters_names: List[str] = field(default_factory=list)
    discrete_parameters_names: List[str] = field(default_factory=list)
    lower_bounds_of_float_parameters: List[float] = field(default_factory=list)
    upper_bounds_of_float_parameters: List[float] = field(default_factory=list)
    discrete_parameters_vals: List[List[Any]] = field(default_factory=list)

    @staticmethod
    def from_parameters_dicts(float_parameters_dict: Optional[Dict[str, List]] = None,
                              discrete_parameters_dict: Optional[Dict[str, List]] = None):
        float_parameters_dict = float_parameters_dict or {}
        discrete_parameters_dict = discrete_parameters_dict or {}

        float_parameters_names = list(float_parameters_dict.keys())
        discrete_parameters_names = list(discrete_parameters_dict.keys())

        lower_bounds_of_float_parameters = [bounds[0] for bounds in float_parameters_dict.values()]
        upper_bounds_of_float_parameters = [bounds[1] for bounds in float_parameters_dict.values()]
        discrete_parameters_vals = [values_set for values_set in discrete_parameters_dict.values()]

        # TODO: Remove - for now IOpt handles only float variables, so we treat discrete parameters as float ones
        float_parameters_names.extend(discrete_parameters_names)
        lower_bounds_of_discrete_parameters = [bounds[0] for bounds in discrete_parameters_dict.values()]
        upper_bounds_of_discrete_parameters = [bounds[1] for bounds in discrete_parameters_dict.values()]
        lower_bounds_of_float_parameters.extend(lower_bounds_of_discrete_parameters)
        upper_bounds_of_float_parameters.extend(upper_bounds_of_discrete_parameters)

        return IOptProblemParameters(float_parameters_names, discrete_parameters_names,
                                     lower_bounds_of_float_parameters,
                                     upper_bounds_of_float_parameters, discrete_parameters_vals)


class GolemProblem(Problem, Generic[DomainGraphForTune]):
    def __init__(self, graph: DomainGraphForTune,
                 objective_evaluate: ObjectiveEvaluate,
                 problem_parameters: IOptProblemParameters):
        super().__init__()
        self.objective_evaluate = objective_evaluate
        self.graph = graph

        self.numberOfObjectives = 1
        self.numberOfConstraints = 0

        self.discreteVariableNames = problem_parameters.discrete_parameters_names
        self.discreteVariableValues = problem_parameters.discrete_parameters_vals
        self.numberOfDiscreteVariables = len(self.discreteVariableNames)

        self.floatVariableNames = problem_parameters.float_parameters_names
        self.lowerBoundOfFloatVariables = problem_parameters.lower_bounds_of_float_parameters
        self.upperBoundOfFloatVariables = problem_parameters.upper_bounds_of_float_parameters
        self.numberOfFloatVariables = len(self.floatVariableNames)

        self._default_metric_value = np.inf

    def Calculate(self, point: Point, functionValue: FunctionValue) -> FunctionValue:
        new_parameters = self.get_parameters_dict_from_iopt_point(point)
        BaseTuner.set_arg_graph(self.graph, new_parameters)
        graph_fitness = self.objective_evaluate(self.graph)
        metric_value = graph_fitness.value if graph_fitness.valid else self._default_metric_value
        functionValue.value = metric_value
        return functionValue

    def get_parameters_dict_from_iopt_point(self, point: Point) -> Dict[str, Any]:
        """Constructs a dict with all hyperparameters """
        float_parameters = dict(zip(self.floatVariableNames, point.floatVariables)) \
            if point.floatVariables is not None else {}
        discrete_parameters = dict(zip(self.discreteVariableNames, point.discreteVariables)) \
            if point.discreteVariables is not None else {}

        # TODO: Remove workaround - for now IOpt handles only float variables, so discrete parameters
        #  are optimized as continuous and we need to round them
        for parameter_name in float_parameters:
            if parameter_name in self.discreteVariableNames:
                float_parameters[parameter_name] = round(float_parameters[parameter_name])

        parameters_dict = {**float_parameters, **discrete_parameters}
        return parameters_dict


class IOptTuner(BaseTuner):
    """
    Base class for hyperparameters optimization based on hyperopt library

    Args:
        objective_evaluate: objective to optimize
        adapter: the function for processing of external object that should be optimized
        iterations: max number of iterations
        search_space: SearchSpace instance
        n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's)
        eps: The accuracy of the solution of the problem. Less value - higher search accuracy, less likely to stop
            prematurely.
        r: Reliability parameter. Higher r is slower convergence, higher probability of finding a global minimum.
        evolvent_density: Density of the evolvent. By default :math:`2^{-10}` on hypercube :math:`[0,1]^N`,
             which means, that the maximum search accuracy is :math:`2^{-10}`.
        eps_r: Parameter that affects the speed of solving the task. epsR = 0 - slow convergence
             to the exact solution, epsR>0 - quick converge to the neighborhood of the solution.
        refine_solution: if true, then the solution will be refined with local search.
        deviation: required improvement (in percent) of a metric to return tuned graph.
            By default, ``deviation=0.05``, which means that tuned graph will be returned
            if it's metric will be at least 0.05% better than the initial.
    """

    def __init__(self, objective_evaluate: ObjectiveEvaluate,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 n_jobs: int = -1,
                 eps: float = 0.01,
                 r: float = 2.0,
                 evolvent_density: int = 10,
                 eps_r: float = 0.001,
                 refine_solution: bool = False,
                 deviation: float = 0.05, **kwargs):
        super().__init__(objective_evaluate,
                         search_space,
                         adapter,
                         iterations=iterations,
                         n_jobs=n_jobs,
                         deviation=deviation, **kwargs)
        self.solver_parameters = SolverParameters(r=np.double(r),
                                                  eps=np.double(eps),
                                                  itersLimit=iterations,
                                                  evolventDensity=evolvent_density,
                                                  epsR=np.double(eps_r),
                                                  refineSolution=refine_solution)

    def tune(self, graph: DomainGraphForTune, show_progress: bool = True) -> DomainGraphForTune:
        graph = self.adapter.adapt(graph)
        problem_parameters, initial_parameters = self._get_parameters_for_tune(graph)

        no_parameters_to_optimize = (not problem_parameters.discrete_parameters_names and
                                     not problem_parameters.float_parameters_names)
        self.init_check(graph)

        if no_parameters_to_optimize:
            self._stop_tuning_with_message(f'Graph "{graph.graph_description}" has no parameters to optimize')
            final_graph = graph
        else:
            if initial_parameters:
                initial_point = Point(**initial_parameters)
                self.solver_parameters.startPoint = initial_point

            problem = GolemProblem(graph, self.objective_evaluate, problem_parameters)
            solver = Solver(problem, parameters=self.solver_parameters)

            if show_progress:
                console_output = ConsoleFullOutputListener(mode='full')
                solver.AddListener(console_output)

            solution = solver.Solve()
            best_point = solution.bestTrials[0].point
            best_parameters = problem.get_parameters_dict_from_iopt_point(best_point)
            final_graph = self.set_arg_graph(graph, best_parameters)

            self.was_tuned = True

        # Validate if optimisation did well
        graph = self.final_check(final_graph)
        final_graph = self.adapter.restore(graph)
        return final_graph

    def _get_parameters_for_tune(self, graph: OptGraph) -> Tuple[IOptProblemParameters, dict]:
        """ Method for defining the search space

        Args:
            graph: graph to be tuned

        Returns:
            parameters_dict: dict with operation names and parameters
            initial_parameters: dict with initial parameters of the graph
        """
        float_parameters_dict = {}
        discrete_parameters_dict = {}
        initial_parameters = {'floatVariables': [], 'discreteVariables': []}
        for node_id, node in enumerate(graph.nodes):
            operation_name = node.name

            # Assign unique prefix for each model hyperparameter
            # label - number of node in the graph
            float_node_parameters, discrete_node_parameters = get_node_parameters_for_iopt(self.search_space,
                                                                                           node_id,
                                                                                           operation_name)

            # Set initial parameters for search
            for parameter, bounds in float_node_parameters.items():
                # If parameter is not set use parameter minimum possible value
                initaial_value = node.parameters.get(parameter) or bounds[0]
                initial_parameters['floatVariables'].append(initaial_value)

            for parameter, bounds in discrete_node_parameters.items():
                # If parameter is not set use parameter minimum possible value
                initaial_value = node.parameters.get(parameter) or bounds[0]
                initial_parameters['discreteVariables'].append(initaial_value)

            float_parameters_dict.update(float_node_parameters)
            discrete_parameters_dict.update(discrete_node_parameters)
        parameters_dict = IOptProblemParameters.from_parameters_dicts(float_parameters_dict, discrete_parameters_dict)
        return parameters_dict, initial_parameters


def get_node_parameters_for_iopt(search_space: SearchSpace, node_id: int, operation_name: str) \
        -> Tuple[Dict[str, List], Dict[str, List]]:
    """
    Method for forming dictionary with hyperparameters of node operation for the ``IOptTuner``

    Args:
        search_space: SearchSpace with parameters per operation
        node_id: number of node in graph.nodes list
        operation_name: name of operation in the node

    Returns:
        float_parameters_dict: dictionary-like structure with labeled float hyperparameters
        and their range per operation
        discrete_parameters_dict: dictionary-like structure with labeled discrete hyperparameters
        and their range per operation
    """
    # Get available parameters for operation
    parameters_dict = search_space.parameters_per_operation.get(operation_name, {})

    discrete_parameters_dict = {}
    float_parameters_dict = {}

    for parameter_name, parameter_properties in parameters_dict.items():
        node_op_parameter_name = get_node_operation_parameter_label(node_id, operation_name, parameter_name)

        parameter_type = parameter_properties.get('type')
        if parameter_type == 'discrete':
            discrete_parameters_dict.update({node_op_parameter_name: parameter_properties
                                            .get('sampling-scope')})
        elif parameter_type == 'continuous':
            float_parameters_dict.update({node_op_parameter_name: parameter_properties
                                         .get('sampling-scope')})

    return float_parameters_dict, discrete_parameters_dict



from random import randint, choices
from typing import Optional, Callable

from golem.core.constants import MAX_GRAPH_GEN_ATTEMPTS
from golem.core.dag.graph import Graph
from golem.core.dag.graph_utils import distance_to_root_level
from golem.core.dag.graph_verifier import GraphVerifier
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.optimisers.opt_node_factory import OptNodeFactory
from golem.core.optimisers.optimization_parameters import GraphRequirements

RandomGraphFactory = Callable[[GraphRequirements, int], Graph]


class RandomGrowthGraphFactory(RandomGraphFactory):
    """ Default realisation of random graph factory. Generates DAG graph using random growth. """
    def __init__(self,
                 verifier: GraphVerifier,
                 node_factory: OptNodeFactory):
        self.node_factory = node_factory
        self.verifier = verifier

    def __call__(self, requirements: GraphRequirements, max_depth: Optional[int] = None) -> OptGraph:
        return random_graph(self.verifier, self.node_factory, requirements, max_depth)


def random_graph(verifier: GraphVerifier,
                 node_factory: OptNodeFactory,
                 requirements: GraphRequirements,
                 max_depth: Optional[int] = None,
                 growth_proba: float = 0.3) -> OptGraph:
    max_depth = max_depth if max_depth else requirements.max_depth
    is_correct_graph = False
    graph = None
    n_iter = 0

    while not is_correct_graph:
        graph = OptGraph()
        graph_root = node_factory.get_node()
        graph.add_node(graph_root)
        if requirements.max_depth > 1:
            graph_growth(graph, graph_root, node_factory, requirements, max_depth, growth_proba)

        is_correct_graph = verifier(graph)
        n_iter += 1
        if n_iter > MAX_GRAPH_GEN_ATTEMPTS:
            raise ValueError(f'Could not generate random graph for {n_iter} '
                             f'iterations with requirements {requirements}')
    return graph


def graph_growth(graph: OptGraph,
                 node_parent: OptNode,
                 node_factory: OptNodeFactory,
                 requirements: GraphRequirements,
                 max_depth: int,
                 growth_proba: float):
    """Function create a graph and links between nodes"""
    offspring_size = randint(requirements.min_arity, requirements.max_arity)

    for offspring_node in range(offspring_size):
        node = node_factory.get_node()
        node_parent.nodes_from.append(node)
        graph.add_node(node)
        height = distance_to_root_level(graph, node)
        is_max_depth_exceeded = height >= max_depth - 1
        if not is_max_depth_exceeded:
            # lower proba of further growth reduces time of graph generation
            if choices([0, 1], weights=[1 - growth_proba, growth_proba])[0]:
                graph_growth(graph, node, node_factory, requirements, max_depth, growth_proba)



import logging
import tracemalloc
from typing import Optional


class MemoryAnalytics:
    is_active = False
    active_session_label = 'main'

    @classmethod
    def start(cls):
        """
        Start memory monitoring session
        """
        cls.is_active = True
        tracemalloc.start()

    @classmethod
    def finish(cls):
        """
        Finish memory monitoring session
        """
        cls.log(additional_info=f'finish',
                logging_level=45)  # message logging level
        tracemalloc.stop()
        cls.is_active = False

    @classmethod
    def get_measures(cls):
        """
        Estimates Python-related system memory consumption in MiB
        :return: current and maximal consumption
        """
        current_memory, max_memory = tracemalloc.get_traced_memory()
        return current_memory / 1024 / 1024, max_memory / 1024 / 1024

    @classmethod
    def log(cls, logger: Optional[logging.LoggerAdapter] = None,
            additional_info: str = 'location', logging_level: int = logging.INFO) -> str:
        """
        Print the message about current and maximal memory consumption to the log or console.
        :param logger: optional logger that should be used in output.
        :param additional_info: label for current location in code.
        :param logging_level: level of the message
        :return: text of the message.
        """
        message = ''
        if cls.is_active:
            memory_consumption = cls.get_measures()
            message = f'Memory consumption for {additional_info} in {cls.active_session_label} session: ' \
                      f'current {round(memory_consumption[0], 1)} MiB, ' \
                      f'max: {round(memory_consumption[1], 1)} MiB'
            if logger is not None:
                logger.log(logging_level, message)
            else:
                print(message)
        return message



from typing import Any, Dict, Type

from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from .. import any_from_json, any_to_json


def parent_operator_to_json(obj: ParentOperator) -> Dict[str, Any]:
    serialized_op = any_to_json(obj)
    serialized_op['parent_individuals'] = [
        parent_ind.uid
        for parent_ind in serialized_op['parent_individuals'] if parent_ind is not None
    ]
    return serialized_op


def parent_operator_from_json(cls: Type[ParentOperator], json_obj: Dict[str, Any]) -> ParentOperator:
    deserialized = any_from_json(cls, json_obj)
    object.__setattr__(deserialized, 'parent_individuals', deserialized.parent_individuals)
    return deserialized



import datetime
from abc import ABC, abstractmethod
from typing import Dict, Iterable, Sequence, Optional, Any, Callable

import numpy as np

from golem.core.optimisers.fitness import is_metric_worse
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.objective.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from .individuals_containers import HallOfFame, ParetoFront

PARETO_MAX_POP_SIZE_MULTIPLIER = 5


class ImprovementWatcher(ABC):
    """Interface that allows to check if optimization progresses or stagnates."""

    @property
    def stagnation_iter_count(self) -> int:
        """Returns number of generations for which any metrics has not improved."""
        raise NotImplementedError()

    @property
    def stagnation_time_duration(self) -> float:
        """Returns time duration for which any metrics has not improved."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_any_improved(self) -> bool:
        """Check if any of the metrics has improved."""
        raise NotImplementedError()

    @abstractmethod
    def is_metric_improved(self, metric_id) -> bool:
        """Check if specified metric has improved."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_quality_improved(self) -> bool:
        """Check if any of the quality metrics has improved."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_complexity_improved(self) -> bool:
        """Check if any of the complexity metrics has improved."""
        raise NotImplementedError()


def _individuals_same(ind1: Individual, ind2: Individual) -> bool:
    return (ind1.fitness == ind2.fitness and
            ind1.native_generation == ind2.native_generation and
            ind1.graph == ind2.graph)


class GenerationKeeper(ImprovementWatcher):
    """Generation keeper that primarily tracks number of generations and stagnation duration.

    Args:
        objective: Objective that specifies metrics and if it's multi objective optimization.
        keep_n_best: How many best individuals to keep from all generations.
         NB: relevant only for single-objective optimization.
        initial_generation: Optional first generation;
         NB: if None then keeper is created in inconsistent state and requires an initial .append().
        similarity_criteria: a function that in the case of multi-objective optimization
         tells the Pareto front whether two individuals are similar, optional.
    """

    def __init__(self,
                 objective: Optional[Objective] = None,
                 keep_n_best: int = 1,
                 initial_generation: PopulationT = None,
                 similarity_criteria: Callable = _individuals_same):
        self._generation_num = 0  # 0 means state before initial generation is added
        self._stagnation_counter = 0  # Initialized in non-stagnated state
        self._stagnation_start_time = datetime.datetime.now()

        self._objective = objective
        self._metrics_improvement: Dict[Any, bool] = {}
        self._reset_metrics_improvement()

        if objective.is_multi_objective:
            self.archive = ParetoFront(maxsize=keep_n_best * PARETO_MAX_POP_SIZE_MULTIPLIER,
                                       similar=similarity_criteria)
        else:
            self.archive = HallOfFame(maxsize=keep_n_best)

        if initial_generation is not None:
            self.append(initial_generation)

    @property
    def stagnation_start_time(self):
        return self._stagnation_start_time

    @property
    def best_individuals(self) -> Sequence[Individual]:
        return self.archive.items

    @property
    def generation_num(self) -> int:
        return self._generation_num

    @property
    def stagnation_iter_count(self) -> int:
        return self._stagnation_counter

    @property
    def stagnation_time_duration(self) -> float:
        return (datetime.datetime.now() - self._stagnation_start_time).seconds / 60

    @property
    def is_any_improved(self) -> bool:
        return any(self._metrics_improvement.values())

    def is_metric_improved(self, metric_id) -> bool:
        return self._metrics_improvement[metric_id]

    @property
    def is_quality_improved(self) -> bool:
        return any(self._metrics_improvement[metric_id]
                   for metric_id in self._objective.quality_metrics)

    @property
    def is_complexity_improved(self) -> bool:
        return any(self._metrics_improvement[metric_id]
                   for metric_id in self._objective.complexity_metrics)

    @property
    def _metric_ids(self) -> Iterable[Any]:
        return self._objective.metric_names

    def append(self, population: PopulationT):
        previous_archive_fitness = self._archive_fitness()
        self.archive.update(population)
        self._update_improvements(previous_archive_fitness)

    def _archive_fitness(self) -> Dict[Any, Sequence[float]]:
        archive_pop_metrics = (ind.fitness.values for ind in self.archive.items)
        archive_fitness_per_metric = zip(*archive_pop_metrics)  # transpose nested array
        archive_fitness_per_metric = dict(zip(self._metric_ids, archive_fitness_per_metric))
        return archive_fitness_per_metric

    def _update_improvements(self, previous_metric_archive):
        self._reset_metrics_improvement()
        current_metric_archive = self._archive_fitness()
        for metric in self._metric_ids:
            # NB: Assuming we perform minimisation, so worst==max
            previous_worst = np.max(previous_metric_archive.get(metric, np.inf))
            current_worst = np.max(current_metric_archive.get(metric, np.inf))
            # archive metric has improved if metric of its worst individual has improved
            if is_metric_worse(previous_worst, current_worst):
                self._metrics_improvement[metric] = True

        self._generation_num += 1  # becomes 1 on first population
        self._stagnation_start_time = datetime.datetime.now() \
            if self.is_any_improved or self._generation_num == 1 else self._stagnation_start_time
        self._stagnation_counter = 0 if self.is_any_improved else self._stagnation_counter + 1

    def _reset_metrics_improvement(self):
        self._metrics_improvement = {metric_id: False for metric_id in self._metric_ids}

    def __str__(self) -> str:
        ff = self._objective.format_fitness
        fitnesses = [ff(ind.fitness) for ind in self.best_individuals]
        return f'{self.archive.__class__.__name__} archive fitness ({len(fitnesses)}): {fitnesses}'



import itertools
from copy import deepcopy
from typing import Any, List, Tuple, Optional

from golem.core.dag.graph_node import descriptive_id_recursive_nodes
from golem.core.dag.graph_utils import distance_to_primary_level


def equivalent_subtree(graph_first: Any, graph_second: Any, with_primary_nodes: bool = False) \
        -> List[Tuple[Any, Any]]:
    """Finds the similar subtrees in two given trees.
    With `with_primary_nodes` primary nodes are considered too.
    Due to a lot of common subgraphs consisted only of single primary nodes, these nodes can be
    not considered with `with_primary_nodes=False`."""

    pairs_list = []
    all_nodes = graph_first.nodes + graph_second.nodes
    all_descriptive_ids = [set(descriptive_id_recursive_nodes(node)) for node in graph_first.nodes] +\
                          [set(descriptive_id_recursive_nodes(node)) for node in graph_second.nodes]
    all_recursive_ids = dict(zip(all_nodes, all_descriptive_ids))
    for node_first in graph_first.nodes:
        for node_second in graph_second.nodes:
            if (node_first, node_second) in pairs_list:
                continue
            equivalent_pairs = structural_equivalent_nodes(node_first=node_first, node_second=node_second,
                                                           recursive_ids=all_recursive_ids)
            pairs_list.extend(equivalent_pairs)

    pairs_list = list(set(pairs_list))
    if with_primary_nodes:
        return pairs_list
    # remove nodes with no children
    result = []
    for pair in pairs_list:
        if len(pair[0].nodes_from) != 0:
            result.append(pair)
    return result


def replace_subtrees(graph_first: Any, graph_second: Any, node_from_first: Any, node_from_second: Any,
                     layer_in_first: int, layer_in_second: int, max_depth: int):
    node_from_graph_first_copy = deepcopy(node_from_first)

    summary_depth = layer_in_first + distance_to_primary_level(node_from_second) + 1
    if summary_depth <= max_depth and summary_depth != 0:
        graph_first.update_subtree(node_from_first, node_from_second)

    summary_depth = layer_in_second + distance_to_primary_level(node_from_first) + 1
    if summary_depth <= max_depth and summary_depth != 0:
        graph_second.update_subtree(node_from_second, node_from_graph_first_copy)


def num_of_parents_in_crossover(num_of_final_inds: int) -> int:
    return num_of_final_inds if not num_of_final_inds % 2 else num_of_final_inds + 1


def filter_duplicates(archive, population) -> List[Any]:
    filtered_archive = []
    for ind in archive.items:
        has_duplicate_in_pop = False
        for pop_ind in population:
            if ind.fitness == pop_ind.fitness:
                has_duplicate_in_pop = True
                break
        if not has_duplicate_in_pop:
            filtered_archive.append(ind)
    return filtered_archive


def structural_equivalent_nodes(node_first: Any,
                                node_second: Any,
                                recursive_ids: Optional[dict] = None,
                                seen: Optional[List[Any]] = None) -> List[Tuple[Any, Any]]:
    """ Returns the list of nodes from which subtrees are structurally equivalent.
    :param node_first: node from first graph from which to start the search.
    :param node_second: node from second graph from which to start the search.
    :param recursive_ids: dict with recursive descriptive id of node with nodes as keys.
    :param seen: list of already visited nodes to avoid infinite recursion.
    Descriptive ids can be obtained with `descriptive_id_recursive_nodes`.
    """

    nodes = []
    is_same_type = type(node_first) == type(node_second)
    seen = seen or []

    if node_first in seen or node_second in seen:
        return []
    seen.append(node_first)
    seen.append(node_second)
    # check if both nodes are primary or secondary
    if hasattr(node_first, 'is_primary') and hasattr(node_second, 'is_primary'):
        is_same_graph_node_type = node_first.is_primary == node_second.is_primary
        is_same_type = is_same_type and is_same_graph_node_type

    for node1_child, node2_child in itertools.product(node_first.nodes_from, node_second.nodes_from):
        nodes_set = structural_equivalent_nodes(node_first=node1_child, node_second=node2_child,
                                                recursive_ids=recursive_ids, seen=seen)
        nodes.extend(nodes_set)
    if is_same_type and len(node_first.nodes_from) == len(node_second.nodes_from) \
            and are_subtrees_the_same(match_set=nodes,
                                      node_first=node_first, node_second=node_second,
                                      recursive_ids=recursive_ids):
        nodes.append((node_first, node_second))
    return nodes


def are_subtrees_the_same(match_set: List[Tuple[Any, Any]],
                          node_first: Any, node_second: Any, recursive_ids: dict = None) -> bool:
    """ Returns `True` if subtrees of specified root nodes are the same, otherwise returns `False`.
    :param match_set: pairs of nodes to checks subtrees from.
    :param node_first: first node from which to compare subtree
    :param node_second: second node from which to compare subtree
    :param recursive_ids: dict with recursive descriptive id of node with nodes as keys.
    Descriptive ids can be obtained with `descriptive_id_recursive_nodes`."""
    matched = []
    if not recursive_ids:
        first_recursive_id = set(descriptive_id_recursive_nodes(node_first))
        second_recursive_id = set(descriptive_id_recursive_nodes(node_second))
    else:
        first_recursive_id = recursive_ids[node_first]
        second_recursive_id = recursive_ids[node_second]

    # 1. Number of exact children must be the same
    # 2. All children from one node must have a match from other node children
    # 3. Protection from cycles when lengths of descriptive ids are the same due to cycles
    if len(node_first.nodes_from) != len(node_second.nodes_from) or \
            len(match_set) == 0 and len(node_first.nodes_from) != 0 or \
            len(first_recursive_id) != len(second_recursive_id):
        return False

    for node, node2 in itertools.product(node_first.nodes_from, node_second.nodes_from):
        if (node, node2) or (node2, node) in match_set:
            matched.append((node, node2))
    if len(matched) >= len(node_first.nodes_from):
        return True
    return False
